"""Foxify Exchange."""

from __future__ import annotations

import asyncio
from decimal import Decimal
import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Self

import keyring
import orjson
from qasync import asyncSlot
from web3 import Account
from web3.types import Gwei

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.exceptions import (
    InvalidOrderSizeError,
    KeyringPasswordNotFoundError,
    TransactionFailedError,
)
from plutus_terminal.core.exchange import helpers as exchange_helpers
from plutus_terminal.core.exchange.base import (
    ExchangeBase,
)
from plutus_terminal.core.exchange.foxify import utils as foxify_utils
from plutus_terminal.core.exchange.foxify.fetcher import FoxifyFetcher
from plutus_terminal.core.exchange.foxify.trader import FoxifyTrader
from plutus_terminal.core.exchange.web3 import web3_utils
from plutus_terminal.core.exchange.web3.cycle_provider import build_cycle_provider
from plutus_terminal.core.types_ import (
    ExchangeType,
    NewAccountInfo,
    PerpsTradeDirection,
    PerpsTradeType,
)
from plutus_terminal.ui.widgets.toast import Toast, ToastType

if TYPE_CHECKING:
    from eth_account.signers.local import LocalAccount

    from plutus_terminal.core.exchange.base import (
        ExchangeFetcherMessageBus,
    )
    from plutus_terminal.core.exchange.types import OrderData, PerpsPosition


LOGGER = logging.getLogger(__name__)


class FoxifyExchange(ExchangeBase):
    """Class to interact with Foxify Exchange."""

    def __init__(self, fetcher_bus: ExchangeFetcherMessageBus) -> None:
        """Initialize shared attributes.

        Args:
            fetcher_bus (ExchangeFetcherMessageBus): ExchangeFetcherMessageBus.
        """
        super().__init__(fetcher_bus=fetcher_bus)
        self.web3_provider = build_cycle_provider("Arbitrum One Trader")
        # Get current account
        keyring_account = CONFIG.current_keyring_account
        keyring_password = keyring.get_password(
            "plutus-terminal",
            str(keyring_account.username),
        )
        if keyring_password is None:
            msg = "Keyring password not found"
            raise KeyringPasswordNotFoundError(msg)
        web3_account: LocalAccount = Account.from_key(orjson.loads(keyring_password)[0])

        self.web3_account = web3_account
        self._pair_prefix = "Crypto."
        self._pair_separator = "/"
        self._quote_symbol = "USD"
        self._pair_suffix = ""
        self._pair_map = self.build_pair_map()
        self._inverted_pair_map = {v: k for k, v in self._pair_map.items()}

    @classmethod
    async def create(cls, fetcher_bus: ExchangeFetcherMessageBus) -> Self:
        """Create class instance and init_async.

        Args:
            fetcher_bus (ExchangeFetcherMessageBus): ExchangeFetcherMessageBus.

        Returns:
            FoxifyExchange: Instance of FoxifyExchange.
        """
        instance = cls(fetcher_bus)
        await instance.init_async()
        return instance

    async def init_async(self) -> None:
        """Initialize async shared attributes."""
        self._trader = await FoxifyTrader.create(
            self._pair_map,
            self.web3_account,
            Gwei(0),
        )
        self._fetcher = await FoxifyFetcher.create(
            self._pair_map,
            self.web3_account.address,
            self.fetcher_bus,
        )
        # Options are being rebuild in Foxify. Disable for now
        # self._options = await FoxifyOptions.create(self.web3_account, Gwei(0))

    @property
    def trader(self) -> FoxifyTrader:
        """Returns: Exchange Trader."""
        return self._trader

    @property
    def fetcher(self) -> FoxifyFetcher:
        """Returns: Exchange Trader."""
        return self._fetcher

    @property
    def available_pairs(self) -> set:
        """Returns set with all available pairs."""
        return set(self._pair_map.values())

    @property
    def pair_map(self) -> dict[str, str]:
        """Returns pair map."""
        return self._pair_map

    @property
    def pair_prefix(self) -> str:
        """Returns pair prefix if any."""
        return self._pair_prefix

    @property
    def pair_separator(self) -> str:
        """Returns the pair separator if any."""
        return self._pair_separator

    @property
    def quote_symbol(self) -> str:
        """Return quote symbol."""
        return self._quote_symbol

    @property
    def pair_suffix(self) -> str:
        """Returns pair suffix."""
        return self._pair_suffix

    @property
    def cached_prices(self) -> dict:
        """Return all prices from cache."""
        return self.fetcher._cached_prices  # noqa: SLF001

    @property
    def account_info(self) -> dict[str, str]:
        """Return info to be added to account info widget."""
        return {
            "Exchange": self.name().capitalize(),
            "Exchange Type": self.exchange_type().name,
            "Wallet": f"{self.web3_account.address[:5]}...{self.web3_account.address[-5:]}",
        }

    @property
    def stable_balance(self) -> Decimal:
        """Return stable balance."""
        return self.fetcher._cached_stable_balance  # noqa: SLF001

    @property
    def min_order_size(self) -> Decimal:
        """Return min trade size."""
        # Could not find this from any contracts only on the front end.
        # TODO: Fix this with the value from the contract
        return Decimal(10)

    # @property
    # def options(self) -> ExchangeOptions:
    #     """Return exchange options."""
    #     return self._options

    @asyncSlot()
    async def is_ready_to_trade(self) -> bool:
        """Check if contracts are approaved.

        Returns:
            bool: True if account is ready to trade.
        """
        router_approved, pos_router_plugin, order_book_plugin = await asyncio.gather(
            foxify_utils.is_stable_approved(
                self.web3_provider,
                self.web3_provider.to_checksum_address(foxify_utils.FOXIFY_ROUTER),
                self.web3_account.address,
            ),
            foxify_utils.is_plugin_approved(
                self.web3_provider,
                foxify_utils.FOXIFY_POSITION_ROUTER,
                self.web3_account.address,
            ),
            foxify_utils.is_plugin_approved(
                self.web3_provider,
                foxify_utils.FOXIFY_ORDER_BOOK,
                self.web3_account.address,
            ),
        )
        return all([router_approved, pos_router_plugin, order_book_plugin])

    @asyncSlot()
    async def approve_for_trading(self) -> None:
        """Approve contracts needed for trading."""
        await foxify_utils.approve_stable(
            self.web3_provider,
            foxify_utils.FOXIFY_ROUTER,
            self.web3_account,
        )
        await foxify_utils.approve_plugin(
            self.web3_provider,
            foxify_utils.FOXIFY_POSITION_ROUTER,
            self.web3_account,
        )
        await foxify_utils.approve_plugin(
            self.web3_provider,
            foxify_utils.FOXIFY_ORDER_BOOK,
            self.web3_account,
        )
        await foxify_utils.ensure_referral(self.web3_provider, self.web3_account)

    async def fetch_prices(self) -> None:
        """Fetch prices in an infinite loop."""
        await self.fetcher.websocket_connect()
        await super().fetch_prices()
        self._async_tasks.append(
            asyncio.create_task(self.fetcher.fetch_positions_funding_rates()),
        )

    def build_pair_map(self) -> dict[str, str]:
        """Build pair map based on foxify pairs.

        Returns:
            dict[str, str]: Pair map.
        """
        with Path.open(Path(__file__).parent.joinpath("pair_map.json")) as f:
            pair_map = json.load(f)

        for pair in pair_map:
            pair_map[pair] = self.format_pair_from_coin(pair_map[pair])

        return pair_map

    @asyncSlot()
    async def create_order(
        self,
        pair: str,
        amount: Decimal,
        trade_direction: PerpsTradeDirection,
        trade_type: PerpsTradeType,
        execution_price: Optional[Decimal] = None,
        take_profit_percent: Optional[float] = None,
        stop_loss_percent: Optional[float] = None,
    ) -> None:
        """Create new order.

        Args:
            pair (str):  Pair to open trade for.
            amount (Decimal): Value in stable to open trade for, this will be multiplied
                by de configured leverage.
            trade_direction (TradeDirection): Trade direction.
            trade_type (TradeType): Trade type.
            execution_price (Optional[Decimal], optional): Execution price.
                If None use current price.
            take_profit_percent (Optional[float], optional): Take profit price.
                If None use CONFIG.take_profit.
            stop_loss_percent (Optional[float], optional): Stop loss price.
                If None use CONFIG.stop_loss.

        Raises:
            InvalidOrderSizeError: If order size is not valid.
            TransactionFailedError: If transaction failed
        """
        if not self.is_valid_order_size(amount):
            msg = (
                f"Invalid order size. Order needs to be: "
                f"less than {self.min_order_size:.2f} and greater than {self.max_order_size:.2f}"
            )
            raise InvalidOrderSizeError(msg)

        toast_id = Toast.show_message(
            f"Creating {trade_type.name} order for {pair}",
            type_=ToastType.WARNING,
        )

        # Get cached price if available otherwise fetch it.
        if execution_price is None and trade_type == PerpsTradeType.MARKET:
            try:
                execution_price = Decimal(self.cached_prices[pair]["price"])
            except KeyError:
                current_price_data = await self.fetcher.fetch_current_price(pair)
                execution_price = Decimal(current_price_data["price"])

            if trade_direction == PerpsTradeDirection.LONG:
                execution_price += execution_price * Decimal(foxify_utils.MAX_SLIPPAGE)
            else:
                execution_price -= execution_price * Decimal(foxify_utils.MAX_SLIPPAGE)
        elif not isinstance(execution_price, Decimal):
            msg = "Invalid execution price."
            raise TypeError(msg)

        acceptable_price = execution_price
        take_profit_execution = exchange_helpers.get_take_profit_target(
            execution_price,
            take_profit_percent,
            trade_direction,
        )
        stop_loss_execution = exchange_helpers.get_stop_loss_target(
            execution_price,
            stop_loss_percent,
            trade_direction,
        )

        size_delta = amount * CONFIG.leverage

        trade_args = foxify_utils.OpenTradingArgs(
            {
                "index_token": self._inverted_pair_map[pair],
                "amount_in": amount,
                "size_delta": size_delta,
                "trade_direction": trade_direction,
                "acceptable_price": acceptable_price,
                "trade_type": trade_type,
                "stop_loss": stop_loss_execution,
                "take_profit": take_profit_execution,
            },
        )
        try:
            trade_result = await self.trader.create_order(trade_args)
        except TransactionFailedError as error:
            Toast.update_message(
                toast_id,
                f"Failed to create order: {error}",
                type_=ToastType.ERROR,
            )
            return

        if isinstance(trade_result, dict):
            msg = "Unexpected error when creating order."
            raise TransactionFailedError(msg)

        await web3_utils.await_receipt_and_report(
            trade_result,
            self.web3_provider,
            "Order Created.",
            "https://arbiscan.io/tx/",
            LOGGER,
            toast_id,
        )

    @asyncSlot()
    async def edit_order(
        self,
        order_data: OrderData,
        new_size_stable: Decimal,
        new_execution_price: Decimal,
    ) -> None:
        """Edit existent order.

        Args:
            order_data (OrderData): Order to edit.
            new_size_stable (Decimal): Value in stable to open trade for, this will be multiplied
                by de configured leverage.
            new_execution_price (Decimal, optional): Execution price.

        Raises:
            TransactionFailedError: If transaction failed
        """
        toast_id = Toast.show_message(
            f"Editing order {order_data['id']}",
            type_=ToastType.WARNING,
        )

        if new_size_stable > order_data["size_stable"]:
            msg = "New size must be smaller/equal than current size."
            raise ValueError(msg)

        order_extra = order_data.get("extra", None)
        if order_extra is None:
            msg = "Missing 'extra' attribute in order_data."
            raise AttributeError(msg)
        order_index = order_extra.get("index", None)
        if order_index is None:
            msg = "Missing 'index' attribute in order_extra."
            raise AttributeError(msg)
        order_trigger_above_threshold = order_extra.get("trigger_above_threshold", None)
        if order_trigger_above_threshold is None:
            msg = "Missing 'trigger_above_threshold' attribute in order_extra."
            raise AttributeError(msg)

        edit_args = foxify_utils.EditOrderArgs(
            {
                "order_index": order_index,
                "size_delta": new_size_stable,
                "acceptable_price": new_execution_price,
                "trigger_above_threshold": order_trigger_above_threshold,
                "trade_type": order_data["order_type"],
            },
        )

        try:
            trade_result = await self.trader.edit_order(edit_args)
        except TransactionFailedError as error:
            Toast.update_message(
                toast_id,
                f"Failed to edit order: {error}",
                type_=ToastType.ERROR,
            )
            return
        if isinstance(trade_result, dict):
            msg = "Unexpected error when creating order."
            raise TransactionFailedError(msg)

        await web3_utils.await_receipt_and_report(
            trade_result,
            self.web3_provider,
            "Order Edited.",
            "https://arbiscan.io/tx/",
            LOGGER,
            toast_id,
        )

    @asyncSlot()
    async def create_reduce_order(
        self,
        pair: str,
        size: Decimal,
        trade_direction: PerpsTradeDirection,
        trade_type: PerpsTradeType,
        execution_price: Optional[Decimal],
    ) -> None:
        """Create new reduce only order.

        Args:
            pair (str):  Pair to open trade for.
            size (Decimal): Value in stable to reduce trade for.
            trade_direction (TradeDirection): Trade direction.
            trade_type (TradeType): Trade type.
            execution_price (Decimal): Execution price.

        Raises:
            TransactionFailedError: If transaction failed
        """
        toast_id = Toast.show_message(
            f"Creating {trade_type.name} order for {pair}",
            type_=ToastType.WARNING,
        )

        # Get cached price if available otherwise fetch it.
        if execution_price is None and trade_type == PerpsTradeType.MARKET:
            try:
                execution_price = Decimal(self.cached_prices[pair]["price"])
            except KeyError:
                current_price_data = await self.fetcher.fetch_current_price(pair)
                execution_price = Decimal(current_price_data["price"])
            if trade_direction == PerpsTradeDirection.LONG:
                execution_price = execution_price + execution_price * Decimal(
                    foxify_utils.MAX_SLIPPAGE,
                )
            else:
                execution_price = execution_price - execution_price * Decimal(
                    foxify_utils.MAX_SLIPPAGE,
                )
        elif not isinstance(execution_price, Decimal):
            msg = "Invalid execution price."
            raise TypeError(msg)

        acceptable_price = execution_price
        trade_args = foxify_utils.ReduceTradingArgs(
            {
                "index_token": self._inverted_pair_map[pair],
                "size_delta": size,
                "trade_direction": trade_direction,
                "acceptable_price": acceptable_price,
                "trade_type": trade_type,
            },
        )

        try:
            trade_result = await self.trader.create_reduce_order(trade_args)
        except TransactionFailedError as error:
            Toast.update_message(
                toast_id,
                f"Failed to create reduce order: {error}",
                type_=ToastType.ERROR,
            )
            return
        except NotImplementedError as error:
            Toast.update_message(
                toast_id,
                f"Reduce order not supported: {error}",
                type_=ToastType.ERROR,
            )
            return

        if isinstance(trade_result, dict):
            msg = "Unexpected error when creating reduce order."
            raise TransactionFailedError(msg)

        await web3_utils.await_receipt_and_report(
            trade_result,
            self.web3_provider,
            "Reduce Order Created.",
            "https://arbiscan.io/tx/",
            LOGGER,
            toast_id,
        )

    @asyncSlot()
    async def cancel_order(self, order_data: OrderData) -> None:
        """Cancel the given order.

        Args:
            order_data  (dict): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.
        """
        toast_id = Toast.show_message(
            f"Canceling {order_data['order_type'].name} order for {order_data['pair']}",
            type_=ToastType.WARNING,
        )

        order_extra = order_data.get("extra", None)
        if order_extra is None:
            msg = "Missing 'extra' attribute in order_data."
            raise AttributeError(msg)
        order_index = order_extra.get("index", None)
        if order_index is None:
            msg = "Missing 'index' attribute in order_extra."
            raise AttributeError(msg)

        cancel_args = foxify_utils.CancelOrderArgs(
            {
                "order_index": int(order_index),
                "reduce_only": order_data["reduce_only"],
            },
        )

        try:
            trade_result = await self.trader.cancel_order(cancel_args)
        except TransactionFailedError as error:
            Toast.update_message(
                toast_id,
                f"Failed to create reduce order: {error}",
                type_=ToastType.ERROR,
            )
            return

        if isinstance(trade_result, dict):
            msg = "Unexpected error when canceling order."
            raise TransactionFailedError(msg)

        await web3_utils.await_receipt_and_report(
            trade_result,
            self.web3_provider,
            "Order Cancelled.",
            "https://arbiscan.io/tx/",
            LOGGER,
            toast_id,
        )

    @asyncSlot()
    async def close_position(self, perp_position: PerpsPosition) -> None:
        """Close position for pair.

        Args:
            perp_position (PerpPosition): Position to close.

        Raises:
            TransactionFailed: If the transaction fails.
        """
        toast_id = Toast.show_message(
            f"Closing position for {perp_position['pair']}",
            type_=ToastType.WARNING,
        )

        # Get cached price if available otherwise fetch it.
        try:
            current_price = Decimal(self.cached_prices[perp_position["pair"]]["price"])
        except KeyError:
            current_price_data = await self.fetcher.fetch_current_price(
                perp_position["pair"],
            )
            current_price = Decimal(current_price_data["price"])

        # Calculate execution price with slippage
        if perp_position["trade_direction"] == PerpsTradeDirection.LONG:
            acceptable_price = current_price - current_price * Decimal(
                foxify_utils.MAX_SLIPPAGE,
            )
        else:
            acceptable_price = current_price + current_price * Decimal(
                foxify_utils.MAX_SLIPPAGE,
            )

        trade_arguments = foxify_utils.CloseTradingArgs(
            {
                "index_token": self._inverted_pair_map[perp_position["pair"]],
                "size_delta": Decimal(perp_position["position_size_stable"]),
                "trade_direction": perp_position["trade_direction"],
                "acceptable_price": acceptable_price,
            },
        )
        try:
            trade_result = await self.trader.close_position(trade_arguments)
        except TransactionFailedError as error:
            Toast.update_message(
                toast_id,
                f"Failed to close position: {error}",
                type_=ToastType.ERROR,
            )
            return

        if isinstance(trade_result, dict):
            msg = "Unexpected error when closing position."
            raise TransactionFailedError(msg)

        await web3_utils.await_receipt_and_report(
            trade_result,
            self.web3_provider,
            "Position Closed.",
            "https://arbiscan.io/tx/",
            LOGGER,
            toast_id,
        )

    # @asyncSlot()
    # async def buy_options_with_strategy(
    #     self, direction: OptionsDirection, value: Decimal, pair: str
    # ) -> None:
    #     """Buy options based on strategy.
    #
    #     Args:
    #         direction (OptionsDirection): Direction to buy the option.
    #         value (Decimal): Max value in stable to spend.
    #         pair (str): Pair to open options for.
    #     """
    #     extracted_data = await self.options.filter_with_strategy(direction, value, pair)
    #     if extracted_data.empty:
    #         Toast.show_message(
    #             "No options avaialble for the given strategy.",
    #             type_=ToastType.ERROR,
    #         )
    #         return
    #
    #     # Store orders ids and amount to buy for each order.
    #     orders_to_buy: OptionsBuyParams = {"orders": {}, "price_id": ""}
    #     available_to_spend = int(value.scaleb(6))
    #
    #     # Set price_id using the first valid oracle priceId
    #     if not orders_to_buy["price_id"]:
    #         orders_to_buy["price_id"] = extracted_data["oracle"].iloc[0]["priceId"]
    #
    #     # Calculate order amounts
    #     for _, row in extracted_data[["orderId", "available"]].iterrows():
    #         if available_to_spend >= row["available"]:
    #             orders_to_buy["orders"][row["orderId"]] = row["available"]
    #         else:
    #             orders_to_buy["orders"][row["orderId"]] = available_to_spend
    #             break
    #         available_to_spend -= orders_to_buy["orders"][row["orderId"]]
    #
    #     await self._options.buy_options(orders_to_buy)

    @staticmethod
    def name() -> str:
        """Return exchange name."""
        return "foxify"

    @staticmethod
    def new_account_info() -> NewAccountInfo:
        """Provide info for new account creation."""
        return NewAccountInfo(
            referral_link="https://perp.foxify.trade/#/trade/?ref=plutus_terminal",
            secrets=["Private Key"],
        )

    @staticmethod
    def exchange_type() -> ExchangeType:
        """Return exchange type."""
        return ExchangeType.DEX

    @staticmethod
    def validate_secrets(secrets: list[str]) -> tuple[bool, str]:
        """Validate private key."""
        try:
            Account.from_key(secrets[0])
        except ValueError:
            return False, "Private key is invalid."
        else:
            return True, "Valid private key."



"""Foxify Funded Exchange."""

from decimal import Decimal

from qasync import asyncSlot
from web3 import Account, HTTPProvider, Web3
from web3.types import Gwei

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.exchange.base import ExchangeFetcherMessageBus
from plutus_terminal.core.exchange.foxify import utils as foxify_utils
from plutus_terminal.core.exchange.foxify.exchange import FoxifyExchange
from plutus_terminal.core.exchange.foxify.funded_fetcher import FoxifyFundedFetcher
from plutus_terminal.core.exchange.foxify.funded_trader import FoxifyFundedTrader
from plutus_terminal.ui.widgets.toast import Toast, ToastType


class FoxifyFundedExchange(FoxifyExchange):
    """Foxify Funded Exchange."""

    def __init__(self, fetcher_bus: ExchangeFetcherMessageBus) -> None:
        """Initialize shared attributes.

        Args:
            fetcher_bus (ExchangeFetcherMessageBus): ExchangeFetcherMessageBus.
        """
        super().__init__(fetcher_bus=fetcher_bus)

    async def init_async(self) -> None:
        """Initialize async shared attributes."""
        self._funded_factory_contract = foxify_utils.build_funded_factory_contract(
            self.web3_provider,
        )

        self._funded_trader_address = self.web3_provider.to_checksum_address(
            await self._funded_factory_contract.functions.traderContracts(
                self.web3_account.address,
            ).call(),
        )

        self._funded_trader_contract = foxify_utils.build_funded_trader_contract(
            self._funded_trader_address,
            self.web3_provider,
        )

        basis_point = await self._funded_trader_contract.functions.BASIS_POINTS().call()
        trader_challenge = await self._funded_trader_contract.functions.traderChallenge().call()
        chanllenge_configs = await self._funded_factory_contract.functions.traderChallengeConfigs(
            trader_challenge,
        ).call()

        min_deposit_stable = chanllenge_configs[14] / 10**foxify_utils.USDC_DECIMAL_PLACES
        self._min_leverage = chanllenge_configs[11] / basis_point
        self._max_leverage = chanllenge_configs[12] / basis_point
        self._min_order_size = Decimal(chanllenge_configs[9] / basis_point) * Decimal(
            min_deposit_stable,
        )
        self._max_order_size = Decimal(chanllenge_configs[10] / basis_point) * Decimal(
            min_deposit_stable,
        )

        self._trader = await FoxifyFundedTrader.create(
            self._pair_map,
            self.web3_account,
            Gwei(0),
        )
        self._fetcher = await FoxifyFundedFetcher.create(
            self._pair_map,
            self._funded_trader_address,
            self.fetcher_bus,
        )

    @property
    def min_order_size(self) -> Decimal:
        """Return min trade size."""
        return self._min_order_size

    @property
    def max_order_size(self) -> Decimal:
        """Return max trade size."""
        return self._max_order_size

    @property
    def account_info(self) -> dict[str, str]:
        """Return info to be added to account info widget."""
        return {
            "Exchange": self.name().capitalize(),
            "Exchange Type": self.exchange_type().name,
            "Wallet": f"{self.web3_account.address[:5]}...{self.web3_account.address[-5:]}",
            "Trader Wallet": f"{self._funded_trader_address[:5]}...{self._funded_trader_address[-5:]}",  # noqa: E501
        }

    @asyncSlot()
    async def is_ready_to_trade(self) -> bool:
        """Check if contracts are approaved.

        Returns:
            bool: True if account is ready to trade.
        """
        # Funded contracts don't need to be approved
        return True

    @asyncSlot()
    async def set_all_leverage(self, leverage: int) -> None:
        """Set leverage for all positions.

        Args:
            leverage (int): Leverage to set.
        """
        toast_id = Toast.show_message(
            f"Leverage set to all pairs: {leverage}x",
            type_=ToastType.SUCCESS,
        )
        if leverage < self._min_leverage:
            Toast.update_message(
                toast_id,
                f"Leverage is too low. Set minimum leverage: {self._min_leverage}x",
                type_=ToastType.WARNING,
            )
            leverage = self._min_leverage
        elif leverage > self._max_leverage:
            Toast.update_message(
                toast_id,
                f"Leverage is too high. Set maximum leverage: {self._max_leverage}x",
                type_=ToastType.WARNING,
            )
            leverage = self._max_leverage

        CONFIG.leverage = leverage

    @asyncSlot()
    async def set_leverage(self, coin: str, leverage: int) -> None:
        """Set leverage for pair.

        Args:
            coin (str): Coin to set leverage for.
            leverage (int): Leverage to set.
        """
        pair = self.format_pair_from_coin(coin)
        toast_id = Toast.show_message(
            f"Leverage of {pair} set to: {leverage}x",
            type_=ToastType.SUCCESS,
        )
        if leverage < self._min_leverage:
            Toast.update_message(
                toast_id,
                f"Leverage of {pair} is too low. Set minimum leverage: {self._min_leverage}x",
                type_=ToastType.WARNING,
            )
            leverage = self._min_leverage
        elif leverage > self._max_leverage:
            Toast.update_message(
                toast_id,
                f"Leverage of {pair} is too high. Set maximum leverage: {self._max_leverage}x",
                type_=ToastType.WARNING,
            )
            leverage = self._max_leverage

        CONFIG.leverage = leverage

    @staticmethod
    def name() -> str:
        """Return exchange name."""
        return "foxify-FUNDED"

    @staticmethod
    def validate_secrets(secrets: list[str]) -> tuple[bool, str]:
        """Validate private key and Funded Trader Contract."""
        try:
            account = Account.from_key(secrets[0])
        except ValueError:
            return False, "Private key is invalid."

        web3_provider = Web3(HTTPProvider("https://arb1.arbitrum.io/rpc"))
        funded_factory_contract = foxify_utils.build_funded_factory_contract(
            web3_provider,  # type: ignore
        )
        trader_address = funded_factory_contract.functions.traderContracts(
            account.address,
        ).call()
        if trader_address == "0x0000000000000000000000000000000000000000":
            return False, "No associated Trader with given private key."

        return True, "Valid private key and Funded Trader Contract."



"""Custom exceptions."""


class OptionsNotAvailableError(Exception):
    """Error raise when Options is not availabe on exchange."""


class TransactionFailedError(Exception):
    """Error raise when transaction failed."""


class KeyringPasswordNotFoundError(Exception):
    """Error raise when password not found in keyring."""


class InvalidOrderSizeError(Exception):
    """Error raise when order size is invalid."""



"""Table to visualize options available to buy."""

from __future__ import annotations

from typing import Any, Optional

import pandas
from PySide6.QtCore import QAbstractTableModel, Qt
from PySide6.QtWidgets import QAbstractItemView, QHeaderView, QTableView, QWidget

HEADER_MAP = {
    "percent": "%",
    "duration": "Time",
    "rate": "Rate",
    "available": "Available",
}

PERCENT_MAP = {
    1000000100000000000: "Up",
    1001000000000000000: "0.10%",
    1002500000000000000: "0.25%",
    1005000000000000000: "0.50%",
    1010000000000000000: "1.00%",
    1030000000000000000: "3.00%",
    1050000000000000000: "5.00%",
    999999900000000000: "Down",
    999000000000000000: "0.10%",
    997500000000000000: "0.25%",
    995000000000000000: "0.50%",
    990000000000000000: "1.00%",
    970000000000000000: "3.00%",
    950000000000000000: "5.00%",
}

DURATION_MAP = {
    15 * 60: "15m",
    30 * 60: "30m",
    60 * 60: "1h",
    2 * 60 * 60: "2h",
    4 * 60 * 60: "4h",
    8 * 60 * 60: "8h",
    24 * 60 * 60: "1d",
}


class OptionsTableModel(QAbstractTableModel):
    """Table model for options available to buy."""

    def __init__(self, data: Optional[pandas.DataFrame] = None) -> None:
        """Initialize table model."""
        QAbstractTableModel.__init__(self)
        self._data = data if data else pandas.DataFrame()
        self.headers_source = list(HEADER_MAP.keys())

    def rowCount(self, parent) -> int:  # type: ignore
        """Get row count."""
        return self._data.shape[0]

    def columnCount(self, parent):  # type: ignore
        """Get column count."""
        return len(self.headers_source)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole) -> Any:  # type: ignore  # noqa: PLR0911
        """Define how data is displayed."""
        current_header = self.headers_source[index.column()]
        if index.isValid():
            if role == Qt.ItemDataRole.DisplayRole:
                if current_header == "percent":
                    return PERCENT_MAP[self._data.iloc[index.row()][current_header]]
                if current_header == "duration":
                    return DURATION_MAP[self._data.iloc[index.row()][current_header]]
                if current_header == "rate":
                    return f"{1 + self._data.iloc[index.row()][current_header] / 10**18:.2f}"
                if current_header == "available":
                    return f"${self._data.iloc[index.row()][current_header] / 10**6:.2f}"

            if role == Qt.ItemDataRole.TextAlignmentRole:
                return Qt.AlignmentFlag.AlignCenter
            return None
        return None

    def headerData(self, section, orientation, role):
        """Define header data."""
        if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            return HEADER_MAP[self.headers_source[section]]
        return None

    def update_data(self, data: pandas.DataFrame) -> None:
        """Update table data."""
        if not self._data.empty and data.equals(self._data):
            return
        self.beginResetModel()
        self._data = data
        self.endResetModel()


class OptionsTableView(QTableView):
    """Table view to display options available to buy."""

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        """Initialize table view."""
        super().__init__(parent=parent)
        self._setup_style()
        self.setMinimumSize(self.sizeHint())

    def _setup_style(self) -> None:
        """Set table style."""
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Stretch)
        self.setAlternatingRowColors(True)
        self.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
        self.setShowGrid(False)
        self.verticalHeader().setVisible(False)



"""Widget to display account info."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from PySide6 import QtWidgets
from PySide6.QtCore import Qt
from PySide6.QtGui import QPixmap
from qasync import asyncSlot

from plutus_terminal.ui.widgets.top_bar_widget import TopBar

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine
    from decimal import Decimal

    from plutus_terminal.core.exchange.base import ExchangeBase


class AccountInfo(QtWidgets.QWidget):
    """Widget to display account info."""

    def __init__(
        self,
        exchange_account_info: dict[str, Any],
        is_ready_to_trade: Callable[[], Coroutine[Any, Any, bool]],
        approve_for_trading: Callable[[], Coroutine[Any, Any, None]],
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent)
        self._exchange_account_info = exchange_account_info
        self._is_ready_to_trade = is_ready_to_trade
        self._approve_for_trading = approve_for_trading

        self.main_layout = QtWidgets.QGridLayout(self)
        self.top_bar = TopBar("Account Info")

        self._frame = QtWidgets.QFrame()
        self._frame_layout = QtWidgets.QGridLayout()
        self._balance_label = QtWidgets.QLabel("Available Balance:")
        self._balance_value = QtWidgets.QLabel("$0.00 USD")
        self._exchange_account_info_layout = QtWidgets.QGridLayout()
        self.approve_btn = QtWidgets.QPushButton("Approve For Trading")

        self._setup_widgets()
        self._setup_layout()

    def _setup_widgets(self) -> None:
        """Configure widgets."""
        self.top_bar.icon.setPixmap(
            QPixmap(":/icons/account_info"),
        )
        self._frame.setObjectName("newsFrameQuote")
        self._balance_value.setAlignment(Qt.AlignmentFlag.AlignRight)
        self._balance_value.setObjectName("subTitle")
        self.approve_btn.setProperty("class", "LONG")
        self.approve_btn.setMinimumHeight(30)
        self.approve_btn.clicked.connect(self._on_approve_for_trading)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.refresh_exchange_account_info(self._exchange_account_info)

    def refresh_exchange_account_info(self, exchange_account_info: dict[str, Any]) -> None:
        """Refresh exchange account info."""
        self._exchange_account_info = exchange_account_info

        # clear exchange_account_info layout
        while self._exchange_account_info_layout.count():
            old_widget = self._exchange_account_info_layout.takeAt(
                self._exchange_account_info_layout.count() - 1,
            ).widget()
            old_widget.deleteLater()

        for label, value in self._exchange_account_info.items():
            label_widget = QtWidgets.QLabel(label)
            value_widget = QtWidgets.QLabel(str(value))
            value_widget.setAlignment(Qt.AlignmentFlag.AlignRight)
            value_widget.setObjectName("subTitle")
            row_count = self._exchange_account_info_layout.rowCount()
            self._exchange_account_info_layout.addWidget(
                label_widget,
                row_count,
                0,
            )
            self._exchange_account_info_layout.addWidget(
                value_widget,
                row_count,
                1,
            )

    def _setup_layout(self) -> None:
        """Configure layout."""
        self.main_layout.addWidget(self.top_bar, 0, 0, 1, 2)

        self._frame_layout.addWidget(self._balance_label, 0, 0)
        self._frame_layout.addWidget(self._balance_value, 0, 1)
        self._frame_layout.addLayout(self._exchange_account_info_layout, 1, 0, 1, 2)
        self._frame.setLayout(self._frame_layout)
        self.main_layout.addWidget(self._frame, 1, 0, 1, 2)
        self.main_layout.addWidget(self.approve_btn, 2, 0, 1, 2)

        self.setLayout(self.main_layout)

    def update_balance(self, balance: Decimal) -> None:
        """Update balance."""
        self._balance_value.setText(f"${balance:.3f} USD")

    @asyncSlot()
    async def set_approve_btn_visibility(self) -> None:
        """Set approve button visibility."""
        if await self._is_ready_to_trade():
            self.approve_btn.setVisible(False)
        else:
            self.approve_btn.setVisible(True)

    @asyncSlot()
    async def _on_approve_for_trading(self) -> None:
        """Approve for trading."""
        await self._approve_for_trading()
        await self.set_approve_btn_visibility()

    def on_new_exchange(self, new_exchange: ExchangeBase) -> None:
        """Update info based on new exchange.

        Args:
            new_exchange (ExchangeBase): New exchangeBase.
        """
        self._is_ready_to_trade = new_exchange.is_ready_to_trade
        self._approve_for_trading = new_exchange.approve_for_trading
        self.refresh_exchange_account_info(new_exchange.account_info)



"""Widget to visualize open orders."""

from __future__ import annotations

from copy import deepcopy
from decimal import Decimal
from functools import partial
from typing import TYPE_CHECKING, Optional

from PySide6.QtCore import (
    QAbstractItemModel,
    QAbstractTableModel,
    QModelIndex,
    Qt,
    Signal,
)
from PySide6.QtGui import QBrush, QColor
from PySide6.QtWidgets import (
    QAbstractItemView,
    QHBoxLayout,
    QHeaderView,
    QPushButton,
    QTableView,
    QWidget,
)
from qasync import asyncSlot

from plutus_terminal.core.exchange.types import (
    OrderData,
    PerpsTradeDirection,
    PerpsTradeType,
)
from plutus_terminal.ui.widgets.manage_order import ManageOrder

if TYPE_CHECKING:
    from collections.abc import Callable

    from plutus_terminal.core.exchange.base import ExchangeBase

HEADER_MAP = {
    "pair": "Pair",
    "trade_direction": "Side",
    "order_type": "Order Type",
    "size_stable": "Size",
    "reduce_only": "Reduce Only",
    "trigger_price": "Trigger Price",
    "buttons": "",
}

FILL_LATER = {"buttons"}


class OrdersTableModel(QAbstractTableModel):
    """Table Model to display open orders."""

    def __init__(
        self,
        format_simple_pair: Callable[[str], str],
        data: Optional[list[OrderData]] = None,
    ) -> None:
        """Initialize shared variables."""
        super().__init__()
        self._data = data if data else []
        self.headers_source = list(HEADER_MAP.keys())
        self.format_simple_pair = format_simple_pair

    def data(self, index, role):  # noqa: C901, PLR0912, PLR0911
        """Define how data is displayed."""
        current_header = self.headers_source[index.column()]

        if current_header not in FILL_LATER:
            value = self._data[index.row()][current_header]
        else:
            value = None

        if role == Qt.ItemDataRole.DisplayRole:
            if value is None:
                return None
            if current_header == "pair":
                return self.format_simple_pair(value)
            if current_header == "trade_direction":
                return value.name
            if current_header == "order_type":
                return value.name.replace("_", " ").title()
            if current_header == "trigger_price":
                if self._data[index.row()]["order_type"] is PerpsTradeType.TRIGGER_TP:
                    if self._data[index.row()]["trade_direction"] is PerpsTradeDirection.LONG:
                        sine = ">"
                    else:
                        sine = "<"
                elif self._data[index.row()]["order_type"] is PerpsTradeType.TRIGGER_SL:
                    if self._data[index.row()]["trade_direction"] is PerpsTradeDirection.LONG:
                        sine = "<"
                    else:
                        sine = ">"
                else:
                    sine = ""
                return f"{sine} ${float(round(value, 4))}"
            if isinstance(value, Decimal):
                return f"${float(round(value, 4))}"
            return str(value).title()

        if role == Qt.ItemDataRole.ForegroundRole and current_header == "trade_direction":
            if value is PerpsTradeDirection.LONG:
                return QBrush(QColor("green"))
            return QBrush(QColor("red"))

        if role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignCenter

        if role == Qt.ItemDataRole.UserRole:
            return self._data[index.row()]
        return None

    def headerData(self, section, orientation, role):
        """Define header data."""
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return HEADER_MAP[self.headers_source[section]]
            return None
        return None

    def rowCount(self, index=None):
        """Define row count."""
        return len(self._data)

    def columnCount(self, index=None):
        """Define column count."""
        return len(HEADER_MAP)

    def update_orders(self, data: list[OrderData]) -> None:
        """Update open orders."""
        # Only reset model if data changed.
        if data == self._data:
            return
        self.beginResetModel()
        self._data = data
        self.endResetModel()

    def on_new_exchange(self, new_exchange: ExchangeBase) -> None:
        """Update info based on new exchange.

        Args:
            new_exchange (ExchangeBase): New exchangeBase.
        """
        self.format_simple_pair = new_exchange.format_simple_pair_from_pair


class OrdersTableView(QTableView):
    """Table view to display open orders."""

    row_clicked = Signal(str)

    def __init__(
        self,
        exchange: ExchangeBase,
        parent: Optional[QWidget] = None,
    ) -> None:
        """Initialize shared viarables."""
        super().__init__(parent)
        self._exchange = exchange
        self.clicked.connect(self.on_row_click)
        self._setup_style()

    def _setup_style(self) -> None:
        """Configure table style."""
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Stretch)
        self.setAlternatingRowColors(True)
        self.verticalHeader().setVisible(False)
        self.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
        self.setShowGrid(False)

    def setModel(self, model: QAbstractItemModel) -> None:
        """Override setModel to add edit and cancel buttons."""
        super().setModel(model)
        model.modelReset.connect(self._add_buttons)

    def _add_buttons(self) -> None:
        """Add edit and cancel buttons for each row."""
        for row in range(self.model().rowCount()):
            buttons = OrderButtons()
            buttons_index = self.model().index(row, list(HEADER_MAP).index("buttons"))
            order_data = buttons_index.data(Qt.ItemDataRole.UserRole)
            buttons.cancel_button.clicked.connect(partial(self.cancel_order, order_data))
            buttons.edit_button.clicked.connect(partial(self._on_edit_order, order_data))
            self.setIndexWidget(buttons_index, buttons)
            self.setRowHeight(row, int(self.sizeHintForRow(row) * 1.1))

        self.horizontalHeader().setSectionResizeMode(
            list(HEADER_MAP).index("buttons"),
            QHeaderView.ResizeMode.Fixed,
        )

        widget = self.indexWidget(
            self.model().index(0, list(HEADER_MAP).index("buttons")),
        )
        if widget is not None:
            self.setColumnWidth(
                list(HEADER_MAP).index("buttons"),
                int(widget.sizeHint().width() * 1.1),
            )

    @asyncSlot(OrderData)
    async def cancel_order(self, order_data: OrderData) -> None:
        """Cancel order."""
        await self._exchange.cancel_order(order_data)

    def _on_edit_order(self, order_data: OrderData) -> None:
        """Handle edit button click.

        Open manage order dialog to edit order.

        Args:
            order_data (OrderData): Order to edit.
        """
        associated_position = self._exchange.get_position_associated_with_order(order_data)
        order_dialog = ManageOrder(
            order_data=deepcopy(order_data),
            exchange=self._exchange,
            associated_position=associated_position,
            parent=self,
        )
        order_dialog.set_edit_mode(True)
        order_dialog.execute_order.connect(
            lambda new_order: self.edit_order(old_order_data=order_data, new_order_data=new_order),
        )
        order_dialog.show()

    @asyncSlot(OrderData, OrderData)
    async def edit_order(self, old_order_data: OrderData, new_order_data: OrderData) -> None:
        """Edit order on exchange.

        Args:
            old_order_data (OrderData): Old order data.
            new_order_data (OrderData): New order data.
        """
        await self._exchange.edit_order(
            order_data=old_order_data,
            new_size_stable=new_order_data["size_stable"],
            new_execution_price=new_order_data["trigger_price"],
        )

    def on_row_click(self, index: QModelIndex) -> None:
        """Handle click on row."""
        self.row_clicked.emit(index.data(Qt.ItemDataRole.UserRole)["pair"])

    def on_new_exchange(self, new_exchange: ExchangeBase) -> None:
        """Update info based on new exchange.

        Args:
            new_exchange (ExchangeBase): New exchangeBase.
        """
        self._exchange = new_exchange


class OrderButtons(QWidget):
    """Button to center."""

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        """Initialize shared attributes."""
        super().__init__(parent)
        layout = QHBoxLayout()
        self.edit_button = QPushButton("Edit")
        self.edit_button.setProperty("class", "gray")
        self.edit_button.setMinimumSize(50, 30)
        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.setProperty("class", "gray")
        self.cancel_button.setMinimumSize(50, 30)
        layout.addWidget(self.edit_button, alignment=Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(self.cancel_button, alignment=Qt.AlignmentFlag.AlignCenter)
        self.setLayout(layout)



"""Widget for Trading Perpetuals."""

from __future__ import annotations

from decimal import Decimal
from functools import partial
from typing import TYPE_CHECKING, Optional

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import Signal
from qasync import asyncSlot

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.exceptions import InvalidOrderSizeError
from plutus_terminal.core.exchange.types import PerpsPosition
from plutus_terminal.core.types_ import PerpsTradeDirection, PerpsTradeType
from plutus_terminal.ui import ui_utils
from plutus_terminal.ui.widgets.decimal_spin_box import DecimalSpinBoxWithButton
from plutus_terminal.ui.widgets.double_spin_button import DoubleSpinBoxWithButton
from plutus_terminal.ui.widgets.toast import Toast, ToastType
from plutus_terminal.ui.widgets.top_bar_widget import TopBar

if TYPE_CHECKING:
    from plutus_terminal.core.exchange.base import ExchangeBase


class PerpsTradeWidget(QtWidgets.QWidget):
    """Widget for Trading Perpetuals."""

    pair_changed = Signal(str)

    def __init__(
        self,
        exchange: ExchangeBase,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent=parent)
        self._exchange = exchange

        self.main_layout = QtWidgets.QGridLayout(self)
        self.top_bar = TopBar("Perps Trade")

        self._quick_buy_grp = QtWidgets.QGroupBox("Quick Market Buy")
        self._quick_buy_layout = QtWidgets.QGridLayout()
        self._long_lowest_btn = QtWidgets.QPushButton()
        self._long_low_btn = QtWidgets.QPushButton()
        self._long_medium_btn = QtWidgets.QPushButton()
        self._long_high_btn = QtWidgets.QPushButton()
        self._long_btns = [
            self._long_lowest_btn,
            self._long_low_btn,
            self._long_medium_btn,
            self._long_high_btn,
        ]
        self._short_lowest_btn = QtWidgets.QPushButton()
        self._short_low_btn = QtWidgets.QPushButton()
        self._short_medium_btn = QtWidgets.QPushButton()
        self._short_high_btn = QtWidgets.QPushButton()
        self._short_btns = [
            self._short_lowest_btn,
            self._short_low_btn,
            self._short_medium_btn,
            self._short_high_btn,
        ]

        self._pair_grp = QtWidgets.QGroupBox("Current Pair")
        self._pair_combo_box = QtWidgets.QComboBox()

        self._leverage_box = QtWidgets.QGroupBox("Current Pair Leverage")
        self._leverage_box_layout = QtWidgets.QGridLayout()
        self._leverage_label = QtWidgets.QLabel("Leverage:")
        self._leverage_spin = QtWidgets.QSpinBox()
        self._leverage_group = QtWidgets.QButtonGroup()
        self._leverage_btn_layout = QtWidgets.QHBoxLayout()

        self._trade_tab = QtWidgets.QTabWidget()
        self._trade_type_market = MarketTradeWidget(
            quote_symbol=self._exchange.quote_symbol,
        )
        self._trade_type_market_layout = QtWidgets.QGridLayout()
        self._trade_type_limit = LimitTradeWidget(
            quote_symbol=self._exchange.quote_symbol,
        )
        self._trade_type_limit_layout = QtWidgets.QGridLayout()

        self._info_frame = QtWidgets.QFrame()
        self._info_layout = QtWidgets.QGridLayout()
        self._leverage_info_label = QtWidgets.QLabel("Leverage:")
        self._leverage_info_value = QtWidgets.QLabel("--")
        self._fees_label = QtWidgets.QLabel("Fees:")
        self._fees_value = QtWidgets.QLabel("--")
        self._liq_price_long_label = QtWidgets.QLabel("Est. Liq. Price LONG:")
        self._liq_price_long_value = QtWidgets.QLabel("--")
        self._liq_price_short_label = QtWidgets.QLabel("Est. Liq. Price SHORT:")
        self._liq_price_short_value = QtWidgets.QLabel("--")

        self._long_button = QtWidgets.QPushButton("Open Long")
        self._short_button = QtWidgets.QPushButton("Open Short")

        self._setup_widgets()
        self._setup_layout()

    def _setup_widgets(self) -> None:  # noqa: PLR0915
        """Configure widgets."""
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.top_bar.icon.setPixmap(
            QtGui.QPixmap(":/icons/dollar_icon"),
        )

        option_keys = (
            "trade_value_lowest",
            "trade_value_low",
            "trade_value_medium",
            "trade_value_high",
        )

        self.update_trade_buttons()
        for index, btn in enumerate(self._long_btns):
            btn.setProperty("class", "LONG")
            btn.setMinimumHeight(25)
            btn.clicked.connect(
                partial(
                    self._handle_quick_trade_click,
                    option_keys[index],
                    PerpsTradeDirection.LONG,
                ),
            )
        for index, btn in enumerate(self._short_btns):
            btn.setProperty("class", "SHORT")
            btn.setMinimumHeight(25)
            btn.clicked.connect(
                partial(
                    self._handle_quick_trade_click,
                    option_keys[index],
                    PerpsTradeDirection.SHORT,
                ),
            )

        self._set_from_exchange()
        self._pair_combo_box.currentTextChanged.connect(
            lambda pair: self.pair_changed.emit(
                f"{self._exchange.pair_prefix}{pair}{self._exchange.pair_suffix}",
            ),
        )

        pair_layout = QtWidgets.QVBoxLayout()
        pair_layout.addWidget(self._pair_combo_box)
        self._pair_grp.setLayout(pair_layout)

        for value in (2, 5, 10, 20, 25, 50):
            button = QtWidgets.QRadioButton(str(value))
            self._leverage_group.addButton(button)
            self._leverage_group.setId(button, value)
            self._leverage_btn_layout.addWidget(button)
        self._leverage_group.buttonClicked.connect(self._set_leverage_button)

        self._leverage_spin.setMinimum(1)
        self._leverage_spin.setMaximum(50)
        self._leverage_spin.setValue(CONFIG.leverage)
        self._leverage_spin.valueChanged.connect(self._set_leverage_spin)
        self._update_leverage_buttons(self._leverage_spin.value())

        self._trade_tab.setObjectName("tradeType")
        self._trade_tab.tabBar().setObjectName("tradeTypeTab")
        self._trade_tab.addTab(self._trade_type_market, "Market")
        self._trade_tab.addTab(self._trade_type_limit, "Limit")
        self._trade_tab.currentChanged.connect(self._update_tab)
        self._trade_tab.setMinimumHeight(self._trade_tab.sizeHint().height() + 5)
        self._trade_tab.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.MinimumExpanding,
            QtWidgets.QSizePolicy.Policy.Minimum,
        )

        self._trade_type_market.amount_changed.connect(
            self._update_info,
        )
        self._trade_type_market.percent_button_clicked.connect(
            self._handle_percent_button_click,
        )
        self._trade_type_limit.price_refresh_btn.clicked.connect(
            self._refresh_limit_price,
        )
        self._trade_type_limit.target_price_box.buttonClicked.connect(
            self._refresh_limit_price,
        )
        self._trade_type_limit.amount_changed.connect(
            self._update_info,
        )
        self._trade_type_limit.percent_button_clicked.connect(
            self._handle_percent_button_click,
        )

        self._info_frame.setObjectName("newsFrameQuote")
        self._leverage_info_value.setText(f"{self._leverage_spin.value()}x")
        self._leverage_info_value.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self._fees_value.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self._liq_price_long_value.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self._liq_price_short_value.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        self._long_button.setProperty("class", "LONG")
        self._long_button.setMinimumHeight(40)
        self._long_button.clicked.connect(
            partial(self._create_order, PerpsTradeDirection.LONG),
        )
        self._short_button.setProperty("class", "SHORT")
        self._short_button.setMinimumHeight(40)
        self._short_button.clicked.connect(
            partial(self._create_order, PerpsTradeDirection.SHORT),
        )

    def _setup_layout(self) -> None:
        """Configure layouts."""
        self.main_layout.addWidget(self.top_bar, 0, 0, 1, 2)

        self._quick_buy_layout.addWidget(self._long_lowest_btn, 0, 0)
        self._quick_buy_layout.addWidget(self._long_low_btn, 0, 1)
        self._quick_buy_layout.addWidget(self._long_medium_btn, 1, 0)
        self._quick_buy_layout.addWidget(self._long_high_btn, 1, 1)
        self._quick_buy_layout.addWidget(self._short_lowest_btn, 0, 2)
        self._quick_buy_layout.addWidget(self._short_low_btn, 0, 3)
        self._quick_buy_layout.addWidget(self._short_medium_btn, 1, 2)
        self._quick_buy_layout.addWidget(self._short_high_btn, 1, 3)
        self._quick_buy_grp.setLayout(self._quick_buy_layout)

        self.main_layout.addWidget(self._pair_grp, 1, 0, 1, 2)
        self.main_layout.addWidget(self._quick_buy_grp, 2, 0, 1, 2)

        self._leverage_box_layout.addWidget(self._leverage_label, 0, 0)
        self._leverage_box_layout.addWidget(self._leverage_spin, 0, 1)
        self._leverage_box_layout.addLayout(self._leverage_btn_layout, 1, 0, 1, 2)
        self._leverage_box.setLayout(self._leverage_box_layout)
        self.main_layout.addWidget(self._leverage_box, 3, 0, 1, 2)

        self.main_layout.addWidget(self._trade_tab, 4, 0, 1, 2)

        self._info_layout.addWidget(self._leverage_info_label, 0, 0)
        self._info_layout.addWidget(self._leverage_info_value, 0, 1)
        self._info_layout.addWidget(self._fees_label, 1, 0)
        self._info_layout.addWidget(self._fees_value, 1, 1)
        self._info_layout.addWidget(self._liq_price_long_label, 2, 0)
        self._info_layout.addWidget(self._liq_price_long_value, 2, 1)
        self._info_layout.addWidget(self._liq_price_short_label, 3, 0)
        self._info_layout.addWidget(self._liq_price_short_value, 3, 1)
        self._info_frame.setLayout(self._info_layout)
        self.main_layout.addWidget(self._info_frame, 5, 0, 1, 2)

        self.main_layout.addWidget(
            self._long_button,
            6,
            0,
            alignment=QtCore.Qt.AlignmentFlag.AlignBottom,
        )
        self.main_layout.addWidget(
            self._short_button,
            6,
            1,
            alignment=QtCore.Qt.AlignmentFlag.AlignBottom,
        )

    def _set_from_exchange(self) -> None:
        """Set data from exchange."""
        # Fill combo box with available pairs
        self._pair_combo_box.clear()
        for pair in sorted(self._exchange.available_pairs):
            self._pair_combo_box.addItem(
                self._exchange.format_simple_pair_from_pair(pair),
                userData=pair,
            )
        # Set current Text to be the default pair from exchange
        default_pair = self._exchange.default_pair
        default_pair = self._exchange.format_simple_pair_from_pair(default_pair)
        self._pair_combo_box.setCurrentText(default_pair)

        self.top_bar.title.setText(f"Persp Trade | {default_pair}")

    @asyncSlot()
    async def _set_leverage_spin(self, leverage_value: int) -> None:
        """Set leverage when spin is changed.

        Update buttons if values matches.

        Args:
            leverage_value (int): Leverage value.
        """
        self._leverage_spin.blockSignals(True)
        self._leverage_spin.setValue(leverage_value)
        self._leverage_spin.blockSignals(False)

        self._update_info()
        self._update_leverage_buttons(leverage_value)

        await self._set_leverage()

    def _update_leverage_buttons(self, leverage_value: int) -> None:
        """Update leverage buttons state based on leverage value."""
        if leverage_value in {2, 5, 10, 20, 25, 50}:
            self._leverage_group.button(int(leverage_value)).setChecked(True)
        else:
            button = self._leverage_group.checkedButton()
            if button:
                self._leverage_group.setExclusive(False)
                button.setChecked(False)
                self._leverage_group.setExclusive(True)

    @asyncSlot()
    async def _set_leverage_button(self, button: QtWidgets.QRadioButton) -> None:
        """Set leverage spin when button is clicked.

        Args:
            button (QtWidgets.QRadioButton): Leverage button clicked.
        """
        await self._set_leverage_spin(self._leverage_group.id(button))

    @asyncSlot()
    async def _set_leverage(self) -> None:
        """Set leverage on exchange for current pair.

        Args:
            leverage_value (int): Leverage value to set.
        """
        leverage_value = self._leverage_spin.value()
        pair = self._pair_combo_box.currentData()
        coin = self._exchange.format_coin_from_pair(pair)
        await self._exchange.set_leverage(coin, leverage_value)

        # In case the levarage was changed due to limits, ensure UI is up to date
        if CONFIG.leverage != leverage_value:
            self._leverage_spin.blockSignals(True)
            self._leverage_spin.setValue(CONFIG.leverage)
            self._update_leverage_buttons(CONFIG.leverage)
            self._leverage_spin.blockSignals(False)

    def _update_info(self) -> None:
        """Update frame info."""
        current_widget = self._trade_tab.currentWidget()
        if not isinstance(current_widget, MarketTradeWidget | LimitTradeWidget):
            return
        amount = current_widget.amount_box.value()
        margin_fee = self._exchange.calculate_margin_fee(
            Decimal(amount) * self._leverage_spin.value(),
        )
        self._fees_value.setText(f"${margin_fee:.3f}")

        leverage_value = self._leverage_spin.value()
        self._leverage_info_value.setText(f"{leverage_value}x")
        self.update_liquidation_info()

    def update_liquidation_info(self) -> None:
        """Update liquidation info."""
        current_widget = self._trade_tab.currentWidget()
        if not isinstance(current_widget, MarketTradeWidget | LimitTradeWidget):
            return

        amount = current_widget.amount_box.value()
        if not amount:
            self._liq_price_long_value.setText("--")
            self._liq_price_short_value.setText("--")
            return

        leverage_value = self._leverage_spin.value()
        pair = self._pair_combo_box.currentData()
        if isinstance(current_widget, LimitTradeWidget):
            open_price = Decimal(current_widget.target_price_box.value())
        else:
            pair_cached = self._exchange.cached_prices.get(pair, None)
            if pair_cached is None:
                return
            open_price = pair_cached["price"]

        long_liq_price = self._exchange.calculate_liquidation_price(
            PerpsPosition(
                {
                    "pair": pair,
                    "id": 0,
                    "position_size_stable": Decimal(amount * leverage_value),
                    "collateral_stable": Decimal(amount),
                    "open_price": open_price,
                    "trade_direction": PerpsTradeDirection.LONG,
                    "leverage": Decimal(leverage_value),
                    "liquidation_price": Decimal(0),
                },
            ),
        )

        minimal_digits = ui_utils.get_minimal_digits(float(long_liq_price), 4)
        self._liq_price_long_value.setText(
            f"<span style='color:rgb(100, 200, 100)'>${long_liq_price:,.{minimal_digits}f}</span>",
        )

        short_liq_price = self._exchange.calculate_liquidation_price(
            PerpsPosition(
                {
                    "pair": pair,
                    "id": 0,
                    "position_size_stable": Decimal(amount * leverage_value),
                    "collateral_stable": Decimal(amount),
                    "open_price": open_price,
                    "trade_direction": PerpsTradeDirection.SHORT,
                    "leverage": Decimal(leverage_value),
                    "liquidation_price": Decimal(0),
                },
            ),
        )
        minimal_digits = ui_utils.get_minimal_digits(float(short_liq_price), 4)
        self._liq_price_short_value.setText(
            f"<span style='color:rgb(255, 100, 100)'>${short_liq_price:,.{minimal_digits}f}</span>",
        )

    @asyncSlot()
    async def _handle_quick_trade_click(
        self,
        option_key: str,
        direction: PerpsTradeDirection,
    ) -> None:
        """Handle quick trade click."""
        amount = getattr(CONFIG, option_key)
        pair = self._pair_combo_box.currentData()
        try:
            await self._exchange.create_order(
                pair,
                amount,
                direction,
                PerpsTradeType.MARKET,
            )
        except InvalidOrderSizeError as error:
            Toast.show_message(
                f"{error}",
                type_=ToastType.ERROR,
            )

    def _handle_percent_button_click(self, button: QtWidgets.QAbstractButton) -> None:
        """Handle percent button click.

        Args:
            button (QtWidgets.QAbstractButton): Percent button clicked.
        """
        current_widget = self._trade_tab.currentWidget()
        if not isinstance(current_widget, MarketTradeWidget | LimitTradeWidget):
            return
        balance = self._exchange.stable_balance
        percentage = Decimal(current_widget.percent_group.id(button) / 100)
        current_widget.amount_box.setValue(balance * percentage)

    @asyncSlot()
    async def _create_order(
        self,
        direction: PerpsTradeDirection,
    ) -> None:
        """Create new order."""
        pair = self._pair_combo_box.currentData()
        current_tab = self._trade_tab.currentWidget()
        if not isinstance(current_tab, LimitTradeWidget) and not isinstance(
            current_tab,
            MarketTradeWidget,
        ):
            return
        amount = Decimal(current_tab.get_amount())
        trade_type = self.get_trade_type()
        # Get None in case the order is Market
        execution_price = (
            Decimal(current_tab.get_target_price())
            if isinstance(current_tab, LimitTradeWidget)
            else None
        )
        stop_loss = Decimal(current_tab.get_stop_loss())
        take_profit = Decimal(current_tab.get_take_profit())
        try:
            await self._exchange.create_order(
                pair,
                amount,
                direction,
                trade_type,
                execution_price,
                take_profit,
                stop_loss,
            )
        except InvalidOrderSizeError as error:
            Toast.show_message(
                f"{error}",
                type_=ToastType.ERROR,
            )

    def get_trade_type(self) -> PerpsTradeType:
        """Get trade type from active tab."""
        current_tab = self._trade_tab.currentWidget()
        if current_tab == self._trade_type_market:
            return PerpsTradeType.MARKET
        return PerpsTradeType.LIMIT

    @asyncSlot()
    async def update_current_pair(self, pair: str) -> None:
        """Update current pair.

        Args:
            pair (str): New pair.
        """
        simplified_pair = self._exchange.format_simple_pair_from_pair(pair)
        self._pair_combo_box.blockSignals(True)
        self._pair_combo_box.setCurrentText(simplified_pair)
        self._pair_combo_box.blockSignals(False)

        # Ensure leverage is set correctly
        coin = self._exchange.format_coin_from_pair(pair)
        await self._exchange.set_leverage(coin, CONFIG.leverage)

        self.top_bar.title.setText(f"Persp Trade | {simplified_pair}")

    def update_trade_buttons(self) -> None:
        """Update trade buttons values."""
        value_map = {
            0: CONFIG.trade_value_lowest,
            1: CONFIG.trade_value_low,
            2: CONFIG.trade_value_medium,
            3: CONFIG.trade_value_high,
        }
        for index, btn in enumerate(self._long_btns):
            btn.setText(f"${value_map[index]}")
        for index, btn in enumerate(self._short_btns):
            btn.setText(f"-${value_map[index]}")

    def _update_tab(self) -> None:
        """Update tab buttons."""
        self._update_info()
        self._trade_type_market.update_button_position()
        self._trade_type_limit.update_button_position()

    def _refresh_limit_price(self) -> None:
        """Refresh limit price."""
        self._trade_type_limit.target_price_box.setValue(
            self._exchange.cached_prices[self._pair_combo_box.currentData()]["price"],
        )

    def on_new_exchange(self, new_exchange: ExchangeBase) -> None:
        """Update info based on new exchange.

        Args:
            new_exchange (ExchangeBase): New exchangeBase.
        """
        self._exchange = new_exchange
        self._set_from_exchange()

    def on_new_account(self) -> None:
        """Update info based on new account."""
        self.blockSignals(True)
        self._leverage_spin.setValue(CONFIG.leverage)
        self._update_leverage_buttons(CONFIG.leverage)
        self.blockSignals(False)


class MarketTradeWidget(QtWidgets.QWidget):
    """Widget to fill a market trade."""

    amount_changed = Signal()
    percent_button_clicked = Signal(QtWidgets.QAbstractButton)

    def __init__(
        self,
        quote_symbol: str,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent=parent)

        self.main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.main_layout)

        self.amount_label = QtWidgets.QLabel("Amount:")
        self.amount_label.setFixedWidth(70)
        self.amount_box = DecimalSpinBoxWithButton(
            button_text=quote_symbol,
        )
        self.amount_box.decimalValueChanged.connect(lambda _: self.amount_changed.emit())
        self.percent_group = QtWidgets.QButtonGroup(self)
        self.percent_group_layout = QtWidgets.QHBoxLayout()
        for value in ("25%", "50%", "75%", "100%"):
            button = QtWidgets.QRadioButton(value)
            self.percent_group.addButton(button)
            self.percent_group.setId(button, int(value[:-1]))
            self.percent_group.buttonClicked.connect(self.percent_button_clicked)
            self.percent_group_layout.addWidget(button)

        amount_layout = QtWidgets.QHBoxLayout()
        amount_layout.addWidget(self.amount_label)
        amount_layout.addWidget(self.amount_box)

        self._take_profit_label = QtWidgets.QLabel("Take Profit:")
        self._take_profit_label.setFixedWidth(70)
        self.take_profit_box = DoubleSpinBoxWithButton(
            button_text="%",
        )
        self.take_profit_box.setDecimals(2)
        take_profit_layout = QtWidgets.QHBoxLayout()
        take_profit_layout.addWidget(self._take_profit_label)
        take_profit_layout.addWidget(self.take_profit_box)

        self._stop_loss_label = QtWidgets.QLabel("Stop Loss:")
        self._stop_loss_label.setFixedWidth(70)
        self.stop_loss_box = DoubleSpinBoxWithButton(
            button_text="%",
        )
        self.stop_loss_box.setDecimals(2)
        stop_loss_layout = QtWidgets.QHBoxLayout()
        stop_loss_layout.addWidget(self._stop_loss_label)
        stop_loss_layout.addWidget(self.stop_loss_box)

        self.main_layout.addLayout(amount_layout)
        self.main_layout.addLayout(self.percent_group_layout)
        separator = QtWidgets.QFrame()
        separator.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        separator.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        separator.setObjectName("divisor")
        self.main_layout.addWidget(separator)
        self.main_layout.addLayout(take_profit_layout)
        self.main_layout.addLayout(stop_loss_layout)
        self.main_layout.addStretch()

    def update_button_position(self) -> None:
        """Update button position."""
        self.amount_box.update_button_position()
        self.take_profit_box.update_button_position()
        self.stop_loss_box.update_button_position()

    def get_amount(self) -> Decimal:
        """Get amount."""
        return self.amount_box.value()

    def get_take_profit(self) -> float:
        """Get take profit."""
        return self.take_profit_box.value()

    def get_stop_loss(self) -> float:
        """Get stop loss."""
        return self.stop_loss_box.value()


class LimitTradeWidget(QtWidgets.QWidget):
    """Widget to fill a limit trade."""

    amount_changed = Signal()
    percent_button_clicked = Signal(QtWidgets.QAbstractButton)

    def __init__(
        self,
        quote_symbol: str,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent=parent)

        self.main_layout = QtWidgets.QGridLayout()
        self.setLayout(self.main_layout)

        self.amount_label = QtWidgets.QLabel("Amount:")
        self.amount_label.setFixedWidth(70)
        self.amount_box = DecimalSpinBoxWithButton(
            button_text=quote_symbol,
        )
        self.amount_box.decimalValueChanged.connect(lambda _: self.amount_changed.emit())
        self.percent_group = QtWidgets.QButtonGroup(self)
        self.percent_group_layout = QtWidgets.QHBoxLayout()
        for value in ("25%", "50%", "75%", "100%"):
            button = QtWidgets.QRadioButton(value)
            self.percent_group.addButton(button)
            self.percent_group.setId(button, int(value[:-1]))
            self.percent_group.buttonClicked.connect(self.percent_button_clicked)
            self.percent_group_layout.addWidget(button)

        self.target_price_label = QtWidgets.QLabel("Price:")
        self.target_price_label.setFixedWidth(70)

        self.target_price_box = DecimalSpinBoxWithButton(
            button_text=quote_symbol,
        )
        self.price_refresh_btn = QtWidgets.QPushButton()
        self.price_refresh_btn.setProperty("class", "borderless")
        self.price_refresh_btn.setIcon(
            QtGui.QIcon(":/icons/focus_icon"),
        )
        self.price_refresh_btn.setFixedWidth(25)

        self._take_profit_label = QtWidgets.QLabel("Take Profit:")
        self._take_profit_label.setFixedWidth(70)
        self.take_profit_box = DoubleSpinBoxWithButton(
            button_text="%",
        )
        self.take_profit_box.setDecimals(2)
        self.take_profit_box.setEnabled(False)

        self._stop_loss_label = QtWidgets.QLabel("Stop Loss:")
        self._stop_loss_label.setFixedWidth(70)
        self.stop_loss_box = DoubleSpinBoxWithButton(
            button_text="%",
        )
        self.stop_loss_box.setDecimals(2)
        self.stop_loss_box.setEnabled(False)

        self.main_layout.addWidget(self.amount_label, 0, 0)
        self.main_layout.addWidget(self.amount_box, 0, 1, 1, 2)
        self.main_layout.addLayout(self.percent_group_layout, 1, 0, 1, 3)
        self.main_layout.addWidget(self.target_price_label, 2, 0)
        self.main_layout.addWidget(self.target_price_box, 2, 1)
        self.main_layout.addWidget(self.price_refresh_btn, 2, 2)
        self.main_layout.setRowStretch(3, 1)
        separator = QtWidgets.QFrame()
        separator.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        separator.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        separator.setObjectName("divisor")
        self.main_layout.addWidget(separator, 3, 0, 1, 3)
        self.main_layout.addWidget(self._take_profit_label, 4, 0)
        self.main_layout.addWidget(self.take_profit_box, 4, 1, 1, 2)
        self.main_layout.addWidget(self._stop_loss_label, 5, 0)
        self.main_layout.addWidget(self.stop_loss_box, 5, 1, 1, 2)
        self.main_layout.rowStretch(6)

    def update_button_position(self) -> None:
        """Update button position."""
        self.amount_box.update_button_position()
        self.target_price_box.update_button_position()
        self.take_profit_box.update_button_position()
        self.stop_loss_box.update_button_position()

    def get_amount(self) -> Decimal:
        """Get amount."""
        return self.amount_box.value()

    def get_target_price(self) -> Decimal:
        """Get price."""
        return self.target_price_box.value()

    def get_take_profit(self) -> float:
        """Get take profit."""
        return self.take_profit_box.value()

    def get_stop_loss(self) -> float:
        """Get stop loss."""
        return self.stop_loss_box.value()



"""Utilities for UI."""

from functools import partial
import math
from typing import Any, Optional, TypeVar

from PySide6.QtCore import QDir
from PySide6.QtWidgets import QWidget

from plutus_terminal.core.exchange.types import PerpsTradeDirection

T = TypeVar("T", bound=QWidget)


def get_minimal_digits(number: float, figures: int) -> int:
    """Get minimal number of digits to show after dot.

    Args:
        number (float): Number to get minimal digits.
        figures (int): Amount of figures after the first non 0.
    """
    if number == 0:
        return 0
    digits = -math.floor(math.log10(abs(number))) + (figures)
    # If 0, figures amount should be showed after dot
    if digits == 0 or digits < figures:
        digits = figures
    return digits


def list_resources_from_prefix(prefix: str) -> list[str]:
    """List all resources of the given prefix."""
    qdir = QDir(":/")
    qdir.setFilter(QDir.Filter.Files | QDir.Filter.Dirs | QDir.Filter.NoDotAndDotDot)
    qdir.setPath(qdir.filePath(prefix))
    return qdir.entryList()


def create_stored_widget(
    widget: type[T],
    storage: dict,
    pair: str,
    trade_direction: PerpsTradeDirection,
    **kwargs: Any,  # noqa: ANN401
) -> T:
    """Create stored widget.

    Args:
        widget (QtWidget): Widget to store.
        storage (dict): Dict to store created widget.
        pair (str): Pair.
        trade_direction (PerpsTradeDirection): Trade direction.
        **kwargs (Any): Additional kwargs.

    Returns:
        QWidget: Stored widget.
    """
    new_widget = widget(**kwargs)
    storage.setdefault(pair, {})
    old_widget = get_stored_widget(storage, pair, trade_direction)
    if old_widget is not None:
        old_widget.deleteLater()
    storage[pair][trade_direction.value] = new_widget
    return new_widget


def get_stored_widget(
    storage: dict,
    pair: str,
    trade_direction: PerpsTradeDirection,
) -> Optional[QWidget]:
    """Get stored widget.

    Args:
        storage (dict): Dict to store created widget.
        pair (str): Pair.
        trade_direction (PerpsTradeDirection): Trade direction.

    Returns:
        QWidget: Stored widget.
    """
    storage.setdefault(pair, {})
    return storage[pair].get(
        trade_direction.value,
        None,
    )



"""Top bar widget."""

from __future__ import annotations

from typing import Optional

from PySide6 import QtWidgets
from PySide6.QtCore import Qt


class TopBar(QtWidgets.QWidget):
    """Top bar widget."""

    def __init__(
        self,
        title: Optional[str] = None,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent)

        self.main_layout = QtWidgets.QVBoxLayout()
        self.bar_layout = QtWidgets.QHBoxLayout()
        self.icon = QtWidgets.QLabel()
        self.title = QtWidgets.QLabel(title)
        self.line = QtWidgets.QFrame()

        self._setup_widgets()
        self._setup_layout()

    def _setup_widgets(self) -> None:
        """Configure widgets."""
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.bar_layout.setContentsMargins(0, 0, 0, 0)

        self.icon.setMinimumHeight(35)
        self.title.setObjectName("title")
        self.title.setAlignment(
            Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
        )

        self.line.setObjectName("divisor")
        self.line.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)

    def _setup_layout(self) -> None:
        """Configure layout."""
        self.bar_layout.addWidget(self.icon)
        self.bar_layout.addWidget(self.title)
        self.bar_layout.addStretch()

        self.setLayout(self.main_layout)
        self.main_layout.addLayout(self.bar_layout)
        self.main_layout.addWidget(self.line)
        self.main_layout.addSpacing(10)

    def add_widget(self, widget: QtWidgets.QWidget) -> None:
        """Add widget to bar_layout."""
        self.bar_layout.addWidget(widget)



"""Foxify Exchange trader."""

from __future__ import annotations

import asyncio
from decimal import Decimal
import logging
from typing import TYPE_CHECKING, Self

from tenacity import before_sleep_log, retry, wait_exponential
from web3.exceptions import ContractLogicError
from web3.types import Gwei, Nonce, Wei

from plutus_terminal.core.exceptions import TransactionFailedError
from plutus_terminal.core.exchange.base import ExchangeTrader
from plutus_terminal.core.exchange.foxify import utils as foxify_utils
from plutus_terminal.core.exchange.types import (
    PerpsTradeDirection,
    PerpsTradeType,
    TradeResults,
)
from plutus_terminal.core.exchange.web3 import web3_utils
from plutus_terminal.core.exchange.web3.cycle_provider import build_cycle_provider
from plutus_terminal.log_utils import log_retry

if TYPE_CHECKING:
    from eth_account.signers.local import LocalAccount


LOGGER = logging.getLogger(__name__)


class FoxifyTrader(ExchangeTrader):
    """Handle trades on Foxify Exchange."""

    def __init__(
        self,
        pair_map: dict[str, str],
        web3_account: LocalAccount,
        extra_gas: Gwei,
    ) -> None:
        """Initialize shared attributes."""
        LOGGER.info("Initialize FoxifyTrader")
        LOGGER.debug("Using web3_account: %s", web3_account)
        LOGGER.debug("Extra gas: %s", extra_gas)
        self.pair_map = pair_map
        self.web3_account = web3_account
        self.extra_gas = extra_gas
        self._receiver_address = web3_account.address
        self.web3_provider = build_cycle_provider("Arbitrum One Trader")
        self._vault_contract = foxify_utils.build_vault_contract(self.web3_provider)
        self._position_router_contract = foxify_utils.build_position_router_contract(
            self.web3_provider,
        )
        self._order_book_contract = foxify_utils.build_order_book_contract(
            self.web3_provider,
        )

    @classmethod
    async def create(
        cls,
        pair_map: dict[str, str],
        web3_account: LocalAccount,
        extra_gas: Gwei,
    ) -> Self:
        """Create class instance and init_async.

        Args:
            pair_map (dict[str, str]): Dict with adress and pair.
            web3_account (LocalAccount): Web3 Account.
            extra_gas (Gwei): Extra gas for all transactions.

        Returns:
            FoxifyTrader: Instance of FoxifyTrader.
        """
        trader = cls(pair_map, web3_account, extra_gas)
        await trader.init_async()
        return trader

    @retry(
        wait=wait_exponential(multiplier=1, min=0.15, max=1),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def init_async(self) -> None:
        """Init async shared attributes."""
        (
            self._price_precision,
            self._position_execution_fee,
            self._order_execution_fee,
        ) = await asyncio.gather(
            self._vault_contract.functions.PRICE_PRECISION().call(),
            self._position_router_contract.functions.minExecutionFee().call(),
            self._order_book_contract.functions.minExecutionFee().call(),
        )

    async def approve_stable(self) -> None:
        """Approve stable token."""

    async def create_order(
        self,
        trade_arguments: foxify_utils.OpenTradingArgs,
    ) -> TradeResults:
        """Create new order.

        Args:
            trade_arguments (foxify.utils.OpenTradingArgs): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.
            NotImplementedError: If the trade direction is not supported.

        Returns:
            TradeResults: Result of the trade.
        """
        LOGGER.info("Opening new position: %s", trade_arguments)
        if trade_arguments["trade_type"] == PerpsTradeType.MARKET:
            return await self._create_market_order(trade_arguments)
        if trade_arguments["trade_type"] == PerpsTradeType.LIMIT:
            return await self._create_limit_order(trade_arguments)
        msg = f"Not supported {trade_arguments['trade_type']}"
        raise NotImplementedError(msg)

    async def _create_market_order(
        self,
        trade_arguments: foxify_utils.OpenTradingArgs,
    ) -> TradeResults:
        """Create a market order.

        Args:
            trade_arguments (foxify.utils.OpenTradingArgs): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        nonce: Nonce = await self.web3_provider.eth.get_transaction_count(
            self.web3_account.address,
        )

        amount_in = int(trade_arguments["amount_in"] * 10**foxify_utils.USDC_DECIMAL_PLACES)
        size_delta = int(trade_arguments["size_delta"] * self._price_precision)
        stop_loss = int(trade_arguments["stop_loss"] * self._price_precision)
        take_profit = int(trade_arguments["take_profit"] * self._price_precision)
        price_bellow = (
            stop_loss
            if trade_arguments["trade_direction"] == PerpsTradeDirection.LONG
            else take_profit
        )
        price_above = (
            take_profit
            if trade_arguments["trade_direction"] == PerpsTradeDirection.LONG
            else stop_loss
        )

        order_execution_data = self.web3_provider.codec.encode(
            types=[
                "address",
                "uint256",
                "uint256",
                "uint256",
                "uint256",
                "uint256",
                "uint256",
                "bool",
            ],
            args=[
                self.web3_provider.to_checksum_address(trade_arguments["index_token"]),
                size_delta if trade_arguments["stop_loss"] != Decimal(0) else 0,
                size_delta if trade_arguments["take_profit"] != Decimal(0) else 0,
                amount_in,
                amount_in,
                price_bellow,
                price_above,
                trade_arguments["trade_direction"].value,
            ],
        )

        try:
            tx = await self._position_router_contract.functions.createIncreasePosition(
                self.web3_provider.to_checksum_address(trade_arguments["index_token"]),
                amount_in,
                0,  # minOut
                size_delta,
                trade_arguments["trade_direction"].value,
                int(trade_arguments["acceptable_price"] * self._price_precision),
                self._position_execution_fee,
                foxify_utils.REFERRAL_CODE,
                order_execution_data,
                foxify_utils.FOXIFY_ROUTER_CALLBACK,
            ).build_transaction(
                {
                    "nonce": nonce,
                    "from": self.web3_account.address,
                    "maxFeePerGas": await web3_utils.estimate_gas_price(
                        self.web3_provider,
                        self.extra_gas,
                    ),
                    "maxPriorityFeePerGas": self.web3_provider.to_wei(
                        self.extra_gas,
                        "gwei",
                    ),
                    "gas": Wei(1000000),
                    "value": self._position_execution_fee,
                },
            )
            tx.update({"gas": await web3_utils.estimate_gas(self.web3_provider, tx)})
            signed_tx = self.web3_account.sign_transaction(tx)
            return await self.web3_provider.eth.send_raw_transaction(
                signed_tx.rawTransaction,
            )
        except (ContractLogicError, ValueError, TypeError) as error:
            LOGGER.exception("Transaction failed")
            raise TransactionFailedError from error

    async def _create_limit_order(
        self,
        trade_arguments: foxify_utils.OpenTradingArgs,
    ) -> TradeResults:
        """Create a limit order.

        Args:
            trade_arguments (foxify.utils.OpenTradingArgs): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.
            NotImplementedError: If the trade direction is not supported.

        Returns:
            TradeResults: Result of the trade.
        """
        nonce: Nonce = await self.web3_provider.eth.get_transaction_count(
            self.web3_account.address,
        )
        try:
            tx = await self._order_book_contract.functions.createIncreaseOrder(
                int(
                    trade_arguments["amount_in"] * 10**foxify_utils.USDC_DECIMAL_PLACES,
                ),
                self.web3_provider.to_checksum_address(trade_arguments["index_token"]),
                int(trade_arguments["size_delta"] * self._price_precision),
                trade_arguments["trade_direction"].value,
                int(
                    trade_arguments["acceptable_price"] * self._price_precision,
                ),  # triggerPrice
                not trade_arguments["trade_direction"].value,
                self._order_execution_fee,
            ).build_transaction(
                {
                    "nonce": nonce,
                    "from": self.web3_account.address,
                    "maxFeePerGas": await web3_utils.estimate_gas_price(
                        self.web3_provider,
                        self.extra_gas,
                    ),
                    "maxPriorityFeePerGas": self.web3_provider.to_wei(
                        self.extra_gas,
                        "gwei",
                    ),
                    "gas": Wei(1000000),
                    "value": self._order_execution_fee,
                },
            )
            tx.update({"gas": await web3_utils.estimate_gas(self.web3_provider, tx)})
            signed_tx = self.web3_account.sign_transaction(tx)
            return await self.web3_provider.eth.send_raw_transaction(
                signed_tx.rawTransaction,
            )
        except (ContractLogicError, ValueError, TypeError) as error:
            LOGGER.exception("Transaction failed")
            raise TransactionFailedError from error

    async def create_reduce_order(
        self,
        trade_arguments: foxify_utils.ReduceTradingArgs,
    ) -> TradeResults:
        """Create new reduce only order.

        Args:
            trade_arguments (foxify.utils.OpenTradingArgs): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.
            NotImplementedError: If the trade direction is not supported.

        Returns:
            TradeResults: Result of the trade.
        """
        if trade_arguments["trade_type"] == PerpsTradeType.MARKET:
            return await self.close_position(trade_arguments)
        if trade_arguments["trade_type"] in (
            PerpsTradeType.TRIGGER_TP,
            PerpsTradeType.TRIGGER_SL,
        ):
            return await self._create_reduce_trigger_order(trade_arguments)
        msg = f"Not supported {trade_arguments['trade_type']}"
        raise NotImplementedError(msg)

    async def _create_reduce_trigger_order(
        self,
        trade_arguments: foxify_utils.ReduceTradingArgs,
    ) -> TradeResults:
        """Create a reduce only trigger order.

        Args:
            trade_arguments (foxify.utils.ReduceTradingArgs): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.
            NotImplementedError: If the trade direction is not supported.

        Returns:
            TradeResults: Result of the trade.
        """
        nonce: Nonce = await self.web3_provider.eth.get_transaction_count(
            self.web3_account.address,
        )

        if trade_arguments["trade_type"] is PerpsTradeType.TRIGGER_SL:
            if trade_arguments["trade_direction"] is PerpsTradeDirection.LONG:
                trigger_above_price = False
            else:
                trigger_above_price = True
        elif trade_arguments["trade_type"] is PerpsTradeType.TRIGGER_TP:
            if trade_arguments["trade_direction"] is PerpsTradeDirection.LONG:
                trigger_above_price = True
            else:
                trigger_above_price = False
        else:
            msg = f"Not supported {trade_arguments['trade_type']}"
            raise NotImplementedError(msg)

        try:
            tx = await self._order_book_contract.functions.createDecreaseOrder(
                self.web3_provider.to_checksum_address(trade_arguments["index_token"]),
                int(trade_arguments["size_delta"] * self._price_precision),
                0,
                trade_arguments["trade_direction"].value,
                int(
                    trade_arguments["acceptable_price"] * self._price_precision,
                ),  # triggerPrice
                trigger_above_price,
            ).build_transaction(
                {
                    "nonce": nonce,
                    "from": self.web3_account.address,
                    "maxFeePerGas": await web3_utils.estimate_gas_price(
                        self.web3_provider,
                        self.extra_gas,
                    ),
                    "maxPriorityFeePerGas": self.web3_provider.to_wei(
                        self.extra_gas,
                        "gwei",
                    ),
                    "gas": Wei(1000000),
                    "value": self._order_execution_fee + 1,
                },
            )
            tx.update({"gas": await web3_utils.estimate_gas(self.web3_provider, tx)})
            signed_tx = self.web3_account.sign_transaction(tx)
            return await self.web3_provider.eth.send_raw_transaction(
                signed_tx.rawTransaction,
            )
        except (ContractLogicError, ValueError, TypeError) as error:
            LOGGER.exception("Transaction failed")
            raise TransactionFailedError from error

    async def close_position(
        self,
        trade_arguments: foxify_utils.CloseTradingArgs,
    ) -> TradeResults:
        """Close the give position.

        Args:
            trade_arguments (foxify.utils.CloseTradingArgs): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.
        """
        LOGGER.info("Closing position: %s", trade_arguments)
        nonce: Nonce = await self.web3_provider.eth.get_transaction_count(
            self.web3_account.address,
        )
        try:
            tx = await self._position_router_contract.functions.createDecreasePosition(
                self.web3_provider.to_checksum_address(trade_arguments["index_token"]),
                0,  # collateralDelta
                int(trade_arguments["size_delta"] * self._price_precision),
                trade_arguments["trade_direction"].value,
                self._receiver_address,
                int(trade_arguments["acceptable_price"] * self._price_precision),
                0,
                self._position_execution_fee,
                "0x0000000000000000000000000000000000000000",
            ).build_transaction(
                {
                    "nonce": nonce,
                    "from": self.web3_account.address,
                    "maxFeePerGas": await web3_utils.estimate_gas_price(
                        self.web3_provider,
                        self.extra_gas,
                    ),
                    "maxPriorityFeePerGas": self.web3_provider.to_wei(
                        self.extra_gas,
                        "gwei",
                    ),
                    "gas": Wei(1000000),
                    "value": self._position_execution_fee,
                },
            )
            tx.update({"gas": await web3_utils.estimate_gas(self.web3_provider, tx)})
            signed_tx = self.web3_account.sign_transaction(tx)
            return await self.web3_provider.eth.send_raw_transaction(
                signed_tx.rawTransaction,
            )
        except (ContractLogicError, ValueError, TypeError) as error:
            LOGGER.exception("Transaction failed")
            raise TransactionFailedError from error

    async def cancel_order(
        self,
        trade_arguments: foxify_utils.CancelOrderArgs,
    ) -> TradeResults:
        """Cancel the given order.

        Args:
            trade_arguments (foxify.utils.CacelOrderArgs): Order to be close.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        LOGGER.info("Canceling order: %s", trade_arguments)
        nonce: Nonce = await self.web3_provider.eth.get_transaction_count(
            self.web3_account.address,
        )
        function_name = (
            "cancelDecreaseOrder" if trade_arguments["reduce_only"] else "cancelIncreaseOrder"
        )
        func = self._order_book_contract.get_function_by_name(function_name)
        try:
            tx = await func(trade_arguments["order_index"]).build_transaction(
                {
                    "nonce": nonce,
                    "from": self.web3_account.address,
                    "maxFeePerGas": await web3_utils.estimate_gas_price(
                        self.web3_provider,
                        self.extra_gas,
                    ),
                    "maxPriorityFeePerGas": self.web3_provider.to_wei(
                        self.extra_gas,
                        "gwei",
                    ),
                    "gas": Wei(1000000),
                },
            )
            tx.update({"gas": await web3_utils.estimate_gas(self.web3_provider, tx)})
            signed_tx = self.web3_account.sign_transaction(tx)
            return await self.web3_provider.eth.send_raw_transaction(
                signed_tx.rawTransaction,
            )
        except (ContractLogicError, ValueError, TypeError) as error:
            LOGGER.exception("Transaction failed")
            raise TransactionFailedError from error

    async def edit_order(
        self,
        trade_arguments: foxify_utils.EditOrderArgs,
    ) -> TradeResults:
        """Update the given order.

        Args:
            trade_arguments (EditOrderArgs): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        if trade_arguments["trade_type"] == PerpsTradeType.LIMIT:
            return await self._edit_limit_order(trade_arguments)
        if trade_arguments["trade_type"] in (
            PerpsTradeType.TRIGGER_TP,
            PerpsTradeType.TRIGGER_SL,
        ):
            return await self._edit_trigger_order(trade_arguments)
        msg = f"Not supported {trade_arguments['trade_type']}"
        raise NotImplementedError(msg)

    async def _edit_limit_order(self, trade_arguments: foxify_utils.EditOrderArgs) -> TradeResults:
        """Edit a limit order.

        Args:
            trade_arguments (foxify.utils.EditOrderArgs): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        LOGGER.info("Editing order: %s", trade_arguments)
        nonce: Nonce = await self.web3_provider.eth.get_transaction_count(
            self.web3_account.address,
        )
        try:
            tx = await self._order_book_contract.functions.updateIncreaseOrder(
                int(trade_arguments["order_index"]),
                int(trade_arguments["size_delta"] * self._price_precision),
                int(trade_arguments["acceptable_price"] * self._price_precision),
                trade_arguments["trigger_above_threshold"],
            ).build_transaction(
                {
                    "nonce": nonce,
                    "from": self.web3_account.address,
                    "maxFeePerGas": await web3_utils.estimate_gas_price(
                        self.web3_provider,
                        self.extra_gas,
                    ),
                    "maxPriorityFeePerGas": self.web3_provider.to_wei(
                        self.extra_gas,
                        "gwei",
                    ),
                    "gas": Wei(1000000),
                },
            )
            tx.update({"gas": await web3_utils.estimate_gas(self.web3_provider, tx)})
            signed_tx = self.web3_account.sign_transaction(tx)
            return await self.web3_provider.eth.send_raw_transaction(
                signed_tx.rawTransaction,
            )
        except (ContractLogicError, ValueError, TypeError) as error:
            LOGGER.exception("Transaction failed")
            raise TransactionFailedError from error

    async def _edit_trigger_order(
        self,
        trade_arguments: foxify_utils.EditOrderArgs,
    ) -> TradeResults:
        """Edit a trigger order.

        Args:
            trade_arguments (foxify.utils.EditOrderArgs): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        LOGGER.info("Editing order: %s", trade_arguments)
        nonce: Nonce = await self.web3_provider.eth.get_transaction_count(
            self.web3_account.address,
        )
        try:
            tx = await self._order_book_contract.functions.updateDecreaseOrder(
                int(trade_arguments["order_index"]),
                0,
                int(trade_arguments["size_delta"] * self._price_precision),
                int(trade_arguments["acceptable_price"] * self._price_precision),
                trade_arguments["trigger_above_threshold"],
            ).build_transaction(
                {
                    "nonce": nonce,
                    "from": self.web3_account.address,
                    "maxFeePerGas": await web3_utils.estimate_gas_price(
                        self.web3_provider,
                        self.extra_gas,
                    ),
                    "maxPriorityFeePerGas": self.web3_provider.to_wei(
                        self.extra_gas,
                        "gwei",
                    ),
                    "gas": Wei(1000000),
                },
            )
            tx.update({"gas": await web3_utils.estimate_gas(self.web3_provider, tx)})
            signed_tx = self.web3_account.sign_transaction(tx)
            return await self.web3_provider.eth.send_raw_transaction(
                signed_tx.rawTransaction,
            )
        except (ContractLogicError, ValueError, TypeError) as error:
            LOGGER.exception("Transaction failed")
            raise TransactionFailedError from error



"""Peewee models."""

from pathlib import Path

import orjson
from peewee import (
    CharField,
    FloatField,
    ForeignKeyField,
    IntegerField,
    Model,
    SqliteDatabase,
    TextField,
)

DATABASE_PATH = Path(__file__).parent.joinpath("plutus_terminal.db")
DATABASE = SqliteDatabase(DATABASE_PATH)


class BaseModel(Model):
    """Base class for Models."""

    class Meta:  # noqa: D106
        database = DATABASE


class KeyringAccount(BaseModel):
    """Model to hold account information."""

    username = CharField(unique=True)
    exchange_type = IntegerField()
    exchange_name = CharField(unique=True)


class TradeConfig(BaseModel):
    """TradeConfig model."""

    account = ForeignKeyField(KeyringAccount, backref="trade_config")
    leverage = IntegerField(default=10)
    stop_loss = FloatField(default=0)
    take_profit = FloatField(default=0)
    trade_value_lowest = IntegerField(default=100)
    trade_value_low = IntegerField(default=250)
    trade_value_medium = IntegerField(default=500)
    trade_value_high = IntegerField(default=1000)


class OptionsConfig(BaseModel):
    """OptionsConfig model."""

    account = ForeignKeyField(KeyringAccount, backref="options_config")
    amount = FloatField(default=100)
    rate_min = FloatField(default=1)
    available_min = FloatField(default=10)
    percent_min = CharField(default="0%")
    percent_max = CharField(default="0%")
    duration_min = IntegerField(default=900)
    duration_max = IntegerField(default=900)
    risk = IntegerField(default=0)


class UserFilter(BaseModel):
    """FilterConfig model."""

    filter_type = IntegerField()
    match_pattern = TextField()
    action_type = IntegerField()
    action_args = TextField()


class GUISettings(BaseModel):
    """Model for GUI settings.

    Expected keys:
        current_account_id
        window_size
        news_show_images
        news_desktop_notifications
        options_show_preview
    """

    key = CharField(unique=True)
    value = TextField()


class Web3RPC(BaseModel):
    """Model for Web3 RPC settings.

    Expected keys:
        rpc_url
    """

    chain_name = CharField(unique=True)
    rpc_urls = TextField()


def create_database() -> None:
    """Create database tables."""
    if DATABASE_PATH.exists():
        return
    with DATABASE:
        DATABASE.create_tables(
            [
                KeyringAccount,
                TradeConfig,
                GUISettings,
                OptionsConfig,
                UserFilter,
                Web3RPC,
            ],
        )

        # Create default GUI settings
        GUISettings.create(key="current_account_id", value=1)

        GUISettings.create(key="news_show_images", value=orjson.dumps(True))
        GUISettings.create(key="news_desktop_notifications", value=orjson.dumps(True))

        GUISettings.create(key="options_show_preview", value=orjson.dumps(False))

        GUISettings.create(key="minimize_to_tray", value=orjson.dumps(True))
        GUISettings.create(key="window_geometry", value=orjson.dumps({}))



"""Exchange helpers."""

from __future__ import annotations

from decimal import Decimal
from typing import Optional

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.exchange.types import PerpsTradeDirection


def get_take_profit_target(
    execution_price: Decimal,
    take_profit_percent: Optional[float],
    trade_direction: PerpsTradeDirection,
) -> Decimal:
    """Get take profit target price."""
    if take_profit_percent is None:
        take_profit_percent = CONFIG.take_profit

    if take_profit_percent == 0:
        return Decimal(0)

    if trade_direction == PerpsTradeDirection.LONG:
        return execution_price * (1 + Decimal(take_profit_percent) / 100)
    return execution_price * (1 - Decimal(take_profit_percent) / 100)


def get_stop_loss_target(
    execution_price: Decimal,
    take_profit_percent: Optional[float],
    trade_direction: PerpsTradeDirection,
) -> Decimal:
    """Get stop loss target price."""
    if take_profit_percent is None:
        take_profit_percent = CONFIG.stop_loss

    if take_profit_percent == 0:
        return Decimal(0)

    if trade_direction == PerpsTradeDirection.LONG:
        return execution_price * (1 - Decimal(take_profit_percent) / 100)
    return execution_price * (1 + Decimal(take_profit_percent) / 100)



"""Web3 utilities functions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from web3.types import TxReceipt, Wei

from plutus_terminal.ui.widgets.toast import Toast, ToastType

if TYPE_CHECKING:
    import logging

    from hexbytes import HexBytes
    from web3 import AsyncWeb3
    from web3.types import Gwei, TxParams


async def estimate_gas_price(web3_provider: AsyncWeb3, extra_gas: Gwei) -> Wei:
    """Estimate transaction gas price.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.
        extra_gas (Gwei): Extra gas.

    Returns:
        Wei: Estimated gas price.
    """
    block_data = await web3_provider.eth.get_block("pending")
    base_fee = block_data["baseFeePerGas"]  # type: ignore
    priority_fee = web3_provider.to_wei(extra_gas, "gwei")
    max_fee = (2 * base_fee) + priority_fee
    return Wei(max_fee)


async def estimate_gas(web3_provider: AsyncWeb3, transaction: TxParams) -> Wei:
    """Estimate amount of gas to use in the transaction.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.
        transaction (TxParams): Transaction params.

    Returns:
        Wei: Estimated gas.
    """
    gas = await web3_provider.eth.estimate_gas(
        {
            "from": transaction["from"],
            "to": transaction["to"],
            "value": transaction["value"],
            "data": transaction["data"],
        },
    )
    gas = int(gas + (gas / 10))
    return Wei(gas)


async def await_receipt_and_report(
    send_txn: HexBytes,
    web3_provider: AsyncWeb3,
    message: str,
    scan_url: str,
    log: logging.Logger,
    toast_id: Optional[bytes] = None,
) -> TxReceipt:
    """Await transaction receipt and report result.

    Args:
        send_txn (HexBytes): Transaction hash.
        web3_provider (AsyncWeb3): Web3 provider.
        message (str): Message to display.
        scan_url (str): Scan url.
        log (logging.Logger): Log instance.
        toast_id (Optional[bytes]): Toast id. If none create one.

    """
    if toast_id is None:
        toast_id = Toast.show_message(
            message=f"Awaiting::{message}...",
            type_=ToastType.WARNING,
            timeout=5000,
        )
    else:
        Toast.update_message(
            toast_id,
            message=f"Awaiting::{message}...",
            type_=ToastType.WARNING,
        )
    tx_receipt = await web3_provider.eth.wait_for_transaction_receipt(send_txn)
    if tx_receipt["status"] == 1:
        log.info(
            "Transaction Sucessfully Sent:: %s Tx:: %s%s",
            message,
            scan_url,
            web3_provider.to_hex(send_txn),
        )
        toast_message = (
            f"Sucessfull::{message} "
            f"<a href='{scan_url}{web3_provider.to_hex(send_txn)}'>TX Link</a>"
        )
        Toast.update_message(
            toast_id,
            message=toast_message,
            type_=ToastType.SUCCESS,
        )
    else:
        log.warning(
            "Transaction Failed:: %s Tx:: %s%s",
            message,
            scan_url,
            web3_provider.to_hex(send_txn),
        )
        toast_message = (
            f"Failed::{message} "
            f"<a href='{scan_url}{web3_provider.to_hex(send_txn)}'>TX Link</a>"
        )
        Toast.update_message(
            toast_id,
            message=toast_message,
            type_=ToastType.ERROR,
        )
    return tx_receipt



"""Widget to control perps configuration."""

from __future__ import annotations

from functools import partial
from typing import Optional

from PySide6 import QtWidgets
from PySide6.QtCore import Signal
from PySide6.QtGui import QPixmap

from plutus_terminal.core.config import CONFIG
from plutus_terminal.ui.widgets.double_spin_button import DoubleSpinBoxWithButton
from plutus_terminal.ui.widgets.toast import Toast, ToastType
from plutus_terminal.ui.widgets.top_bar_widget import TopBar


class PerpsConfig(QtWidgets.QWidget):
    """Widget to control perps configuration."""

    updated_trade_values = Signal()
    leverage_changed = Signal(int)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Initialize widget."""
        super().__init__(parent=parent)

        self.main_layout = QtWidgets.QVBoxLayout(self)

        self.top_bar = TopBar("Perps Config")
        self._auto_tp_sl_box = QtWidgets.QGroupBox("Auto TP/SL")
        self._auto_tp_sl_box_layout = QtWidgets.QGridLayout()
        self._tp_label = QtWidgets.QLabel("Take Profit:")
        self._tp_spin = DoubleSpinBoxWithButton(button_text="%")
        self._sl_label = QtWidgets.QLabel("Stop Loss:")
        self._sl_spin = DoubleSpinBoxWithButton(button_text="%")
        self._tp_sl_update = QtWidgets.QPushButton("Update TP/SL")
        self._advanced_bar = TopBar("Advanced Config")
        self._advanced_box_layout = QtWidgets.QVBoxLayout()
        self._trade_values_box = QtWidgets.QGroupBox("Quick Buy Values")
        self._trade_values_layout = QtWidgets.QGridLayout()
        self._trade_lowest_label = QtWidgets.QLabel("Lowest:")
        self._trade_lowest_spin = QtWidgets.QSpinBox()
        self._trade_low_label = QtWidgets.QLabel("Low:")
        self._trade_low_spin = QtWidgets.QSpinBox()
        self._trade_med_label = QtWidgets.QLabel("Med:")
        self._trade_med_spin = QtWidgets.QSpinBox()
        self._trade_high_label = QtWidgets.QLabel("High:")
        self._trade_high_spin = QtWidgets.QSpinBox()
        self._trade_values_update = QtWidgets.QPushButton("Change Quick Buy Values")
        self._leverage_box = QtWidgets.QGroupBox("Leverage Config")
        self._leverage_box_layout = QtWidgets.QGridLayout()
        self._leverage_label = QtWidgets.QLabel("Leverage:")
        self._leverage_spin = QtWidgets.QSpinBox()
        self._leverage_group = QtWidgets.QButtonGroup()
        self._leverage_layout = QtWidgets.QHBoxLayout()
        self._leverage_set_button = QtWidgets.QPushButton("Set Leverage for All Pairs")

        self._spin_config_map: dict[
            QtWidgets.QSpinBox | QtWidgets.QDoubleSpinBox,
            str,
        ] = {
            self._tp_spin: "take_profit",
            self._sl_spin: "stop_loss",
            self._trade_lowest_spin: "trade_value_lowest",
            self._trade_low_spin: "trade_value_low",
            self._trade_med_spin: "trade_value_medium",
            self._trade_high_spin: "trade_value_high",
        }

        self._setup_widgets()
        self._setup_layout()

    def _setup_widgets(self) -> None:
        """Configure widgets."""
        self.main_layout.setContentsMargins(0, 0, 0, 0)

        self.top_bar.icon.setPixmap(QPixmap(":/icons/perps_config_icon"))

        self._tp_spin.setValue(CONFIG.take_profit)
        self._tp_spin.setMinimum(0)
        self._tp_spin.setMaximum(100)
        self._tp_spin.setDecimals(2)

        self._sl_spin.setValue(CONFIG.stop_loss)
        self._sl_spin.setMinimum(0)
        self._sl_spin.setMaximum(100)
        self._sl_spin.setDecimals(2)

        self._tp_sl_update.setMinimumHeight(35)
        self._tp_sl_update.clicked.connect(self._update_tp_sl)

        self._trade_lowest_spin.setMinimum(1)
        self._trade_lowest_spin.setMaximum(100_000_000)
        self._trade_lowest_spin.setValue(CONFIG.trade_value_lowest)

        self._trade_low_spin.setMinimum(1)
        self._trade_low_spin.setMaximum(100_000_000)
        self._trade_low_spin.setValue(CONFIG.trade_value_low)

        self._trade_med_spin.setMinimum(1)
        self._trade_med_spin.setMaximum(100_000_000)
        self._trade_med_spin.setValue(CONFIG.trade_value_medium)

        self._trade_high_spin.setMinimum(1)
        self._trade_high_spin.setMaximum(100_000_000)
        self._trade_high_spin.setValue(CONFIG.trade_value_high)

        self._trade_values_update.setMinimumHeight(35)
        self._trade_values_update.clicked.connect(self._update_trade_values)

        for value in (2, 5, 10, 20, 25, 50):
            button = QtWidgets.QRadioButton(str(value))
            self._leverage_group.addButton(button)
            self._leverage_group.setId(button, value)
            self._leverage_layout.addWidget(button)
        self._leverage_group.buttonClicked.connect(self._set_leverage_button)

        self._leverage_spin.setMinimum(1)
        self._leverage_spin.setMaximum(50)
        self._leverage_spin.valueChanged.connect(self._update_leverage_buttons)
        self._leverage_spin.setValue(CONFIG.leverage)

        self._leverage_set_button.setMinimumHeight(35)
        self._leverage_set_button.clicked.connect(self._set_leverage)

    def _setup_layout(self) -> None:
        """Configure layout."""
        self.main_layout.addWidget(self.top_bar)

        self._auto_tp_sl_box_layout.addWidget(self._tp_label, 0, 0)
        self._auto_tp_sl_box_layout.addWidget(self._tp_spin, 0, 1)
        self._auto_tp_sl_box_layout.addWidget(self._sl_label, 1, 0)
        self._auto_tp_sl_box_layout.addWidget(self._sl_spin, 1, 1)
        self._auto_tp_sl_box_layout.addWidget(self._tp_sl_update, 2, 0, 1, 2)
        self._auto_tp_sl_box.setLayout(self._auto_tp_sl_box_layout)
        self.main_layout.addWidget(self._auto_tp_sl_box)

        self._trade_values_layout.addWidget(self._trade_lowest_label, 0, 0)
        self._trade_values_layout.addWidget(self._trade_lowest_spin, 0, 1)
        self._trade_values_layout.addWidget(self._trade_low_label, 1, 0)
        self._trade_values_layout.addWidget(self._trade_low_spin, 1, 1)
        self._trade_values_layout.addWidget(self._trade_med_label, 2, 0)
        self._trade_values_layout.addWidget(self._trade_med_spin, 2, 1)
        self._trade_values_layout.addWidget(self._trade_high_label, 3, 0)
        self._trade_values_layout.addWidget(self._trade_high_spin, 3, 1)
        self._trade_values_layout.addWidget(self._trade_values_update, 4, 0, 1, 2)
        self._trade_values_box.setLayout(self._trade_values_layout)
        self._advanced_box_layout.addWidget(self._trade_values_box)
        self._leverage_box_layout.addWidget(self._leverage_label, 0, 0)
        self._leverage_box_layout.addWidget(self._leverage_spin, 0, 1)
        self._leverage_box_layout.addLayout(self._leverage_layout, 1, 0, 1, 2)
        self._leverage_box_layout.addWidget(self._leverage_set_button, 2, 0, 1, 2)
        self._leverage_box.setLayout(self._leverage_box_layout)
        self._advanced_box_layout.addWidget(self._leverage_box)
        self._advanced_bar.main_layout.addLayout(self._advanced_box_layout)

        self.main_layout.addWidget(self._advanced_bar)
        self.main_layout.addStretch()

    def _set_leverage_spin(self, leverage_value: int) -> None:
        """Set leverage when spin is changed.

        Update buttons if values matches.

        Args:
            leverage_value (int): Leverage value.
        """
        self._leverage_spin.blockSignals(True)
        self._leverage_spin.setValue(leverage_value)
        self._leverage_spin.blockSignals(False)

    def _update_leverage_buttons(self, leverage_value: int) -> None:
        """Update leverage buttons state based on leverage value."""
        if leverage_value in {2, 5, 10, 20, 25, 50}:
            self._leverage_group.button(int(leverage_value)).setChecked(True)
        else:
            button = self._leverage_group.checkedButton()
            if button:
                self._leverage_group.setExclusive(False)
                button.setChecked(False)
                self._leverage_group.setExclusive(True)

    def _set_leverage_button(self, button: QtWidgets.QRadioButton) -> None:
        """Set leverage spin when button is clicked.

        Args:
            button (QtWidgets.QRadioButton): Leverage button clicked.
        """
        self._set_leverage_spin(self._leverage_group.id(button))

    def _set_leverage(self) -> None:
        """Set leverage on exchange.

        Args:
            leverage_value (int): Leverage value to set.
        """
        leverage_value = self._leverage_spin.value()
        self.leverage_changed.emit(leverage_value)

    def _update_trade_values(self) -> None:
        """Update trade values."""
        CONFIG.trade_value_lowest = self._trade_lowest_spin.value()
        CONFIG.trade_value_low = self._trade_low_spin.value()
        CONFIG.trade_value_medium = self._trade_med_spin.value()
        CONFIG.trade_value_high = self._trade_high_spin.value()
        self.updated_trade_values.emit()

    def _update_tp_sl(self) -> None:
        """Update take profit and stop loss."""
        CONFIG.take_profit = self._tp_spin.value()
        CONFIG.stop_loss = self._sl_spin.value()
        Toast.show_message("TP/SL values updated", type_=ToastType.SUCCESS)

    def on_new_account(self) -> None:
        """Update info based on new account."""
        self.blockSignals(True)
        # Update spin box values
        for spin, attr in self._spin_config_map.items():
            spin.blockSignals(True)
            spin.setValue(getattr(CONFIG, attr))
            spin.blockSignals(False)

        self._set_leverage_spin(CONFIG.leverage)
        self.blockSignals(False)



"""Plutus terminal main window."""

from __future__ import annotations

import asyncio
from pathlib import Path
from typing import TYPE_CHECKING

import pandas
from PySide6.QtCore import Qt
from PySide6.QtGui import QCloseEvent, QKeySequence, QPixmap, QShortcut
from PySide6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QMainWindow,
    QScrollArea,
    QSizePolicy,
    QSplitter,
    QVBoxLayout,
    QWidget,
)
from qasync import asyncSlot

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.exchange.base import (
    LOGGER,
    ExchangeBase,
    ExchangeFetcherMessageBus,
)
from plutus_terminal.core.exchange.valid_exchanges import VALID_EXCHANGES
from plutus_terminal.core.news.base import NewsMessageBus
from plutus_terminal.core.news.news_manager import NewsManager
from plutus_terminal.ui import ui_utils
from plutus_terminal.ui.widgets.account_info import AccountInfo
from plutus_terminal.ui.widgets.config import ConfigDialog
from plutus_terminal.ui.widgets.news_list import NewsList
from plutus_terminal.ui.widgets.options_widget import OptionsWidget
from plutus_terminal.ui.widgets.perps_trade import PerpsTradeWidget
from plutus_terminal.ui.widgets.toast import Toast, ToastType
from plutus_terminal.ui.widgets.trade_table import TradeTable
from plutus_terminal.ui.widgets.trading_chart import TradingChart
from plutus_terminal.ui.widgets.user_top_bar import UserTopBar

if TYPE_CHECKING:
    from plutus_terminal.core.db.models import KeyringAccount


# TODO: To remove later
def reload_style() -> None:  # noqa: D103
    relative_path = Path(__file__).parent
    with Path.open(relative_path.joinpath("style.qss")) as f:
        QApplication.instance().setStyleSheet(f.read())


class PlutusTerminal(QMainWindow):
    """Plutus terminal main window."""

    def __init__(self) -> None:
        """Initialize shared variables."""
        super().__init__()
        self._fetcher_message_bus = ExchangeFetcherMessageBus()
        self._news_message_bus = NewsMessageBus()
        self._async_tasks = []

        self.main_layout = QVBoxLayout()
        self.main_widget = QWidget()
        self._work_area_layout = QHBoxLayout()
        self._left_splitter = QSplitter(Qt.Orientation.Vertical)
        self._right_layout = QVBoxLayout()
        self._right_scroll = QScrollArea()

        # Declare classes for visibility
        self._current_exchange: ExchangeBase
        self._current_pair: str
        self._chart: TradingChart
        self._trade_table: TradeTable
        self._account_info: AccountInfo
        self._perps_trade: PerpsTradeWidget
        self._config_dialog: ConfigDialog
        self._user_top_bar: UserTopBar
        self._news_manager: NewsManager
        self._news_list: NewsList
        self._options_widget: OptionsWidget

        # TODO: To remove later
        self.shortcut = QShortcut(QKeySequence("F1"), self)
        self.shortcut.activated.connect(reload_style)

    async def init_async(self) -> None:
        """Initialize async shared variables."""
        self._config_dialog = ConfigDialog(self)

        self._user_top_bar = UserTopBar(self._config_dialog)
        self._user_top_bar.account_picker.account_changed.connect(self.change_account)

        # Start current exchange loops
        keyring_account: KeyringAccount = self._user_top_bar.account_picker.currentData()
        self._current_exchange = await VALID_EXCHANGES[str(keyring_account.exchange_name)].create(
            self._fetcher_message_bus,
        )
        self._async_tasks.append(asyncio.create_task(self._current_exchange.fetch_prices()))
        self._current_pair = self._current_exchange.default_pair

        self._chart = TradingChart(
            self._current_exchange.available_pairs,
            self._current_exchange.format_simple_pair_from_pair,
        )

        # Init open trades widget
        self._trade_table = TradeTable(self._current_exchange)

        # Init account info widget
        self._account_info = AccountInfo(
            self._current_exchange.account_info,
            self._current_exchange.is_ready_to_trade,
            self._current_exchange.approve_for_trading,
            parent=self,
        )

        # Init perps trading
        self._perps_trade = PerpsTradeWidget(self._current_exchange)

        # Init news manager
        self._news_manager = NewsManager(self._news_message_bus)
        self._async_tasks.append(asyncio.create_task(self._news_manager.fetch_news()))
        self._news_list = NewsList(self._current_exchange)

        # Init options Widget
        self._options_widget = OptionsWidget(self._current_exchange)
        if not self._current_exchange.has_options():
            self._options_widget.setEnabled(False)
            self._options_widget.setVisible(False)

        await self._setup_widgets()
        self._setup_layout()

        self._exchange_update_affected = [
            widget for widget in self.findChildren(QWidget) if hasattr(widget, "on_new_exchange")
        ]
        self._account_update_affected = [
            widget for widget in self.findChildren(QWidget) if hasattr(widget, "on_new_account")
        ]

    async def _setup_widgets(self) -> None:
        """Configure widgets."""
        self.main_widget.setLayout(self.main_layout)
        self.setWindowTitle("Plutus Terminal")
        self.setWindowIcon(QPixmap(":/icons/plutus_icon"))

        # Set chart data and connect signals
        await self._set_chart_timeframe("1")
        self._chart.current_pair = self._current_pair
        self._chart.timeframe_signal.connect(self._set_chart_timeframe)
        self._fetcher_message_bus.subscribed_prices_signal.connect(
            self._chart.update_chart_tick,
        )
        self._chart.pair_changed.connect(self._change_current_pair)
        self._fetcher_message_bus.positions_signal.connect(
            self._chart.draw_positions,
        )
        self._fetcher_message_bus.orders_signal.connect(self._chart.draw_orders)

        # Configure config dialog
        self._config_dialog.updated_trade_values.connect(
            self._update_quick_trade_values,
        )
        self._config_dialog.leverage_changed.connect(
            self._current_exchange.set_all_leverage,
        )

        # Configure account info
        await self._account_info.set_approve_btn_visibility()
        self._fetcher_message_bus.balance_signal.connect(self._account_info.update_balance)

        # Configure Perps Trade
        self._perps_trade.pair_changed.connect(self._change_current_pair)
        self._fetcher_message_bus.subscribed_prices_signal.connect(
            self._perps_trade.update_liquidation_info,
        )

        # Connect signals for open traders
        self._fetcher_message_bus.positions_signal.connect(
            self._trade_table.update_positions,
        )
        self._fetcher_message_bus.orders_signal.connect(self._trade_table.update_orders)
        self._fetcher_message_bus.subscribed_prices_signal.connect(
            self._trade_table.update_prices,
        )
        self._trade_table.pair_clicked.connect(self._change_current_pair)

        self._news_list.pair_clicked.connect(self._change_current_pair)
        self._news_list.refresh_news.connect(self._fill_news_list)
        await self._fill_news_list()
        self._news_message_bus.news_signal.connect(self._news_list.add_news)

        self._right_scroll.setSizePolicy(
            QSizePolicy.Policy.Fixed,
            QSizePolicy.Policy.MinimumExpanding,
        )
        self._right_scroll.setWidgetResizable(True)
        self._right_scroll.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff,
        )

        self._options_widget.setSizePolicy(
            QSizePolicy.Policy.Fixed,
            QSizePolicy.Policy.Fixed,
        )
        self._options_widget.setMinimumWidth(400)

    def _setup_layout(self) -> None:
        """Organize layouts."""
        self.main_layout.addWidget(self._user_top_bar)

        self._left_splitter.addWidget(self._chart)
        self._left_splitter.addWidget(self._trade_table)
        self._work_area_layout.addWidget(self._left_splitter)
        self._work_area_layout.addWidget(self._news_list)

        self._right_layout.addWidget(self._account_info)
        self._right_layout.addWidget(self._perps_trade)
        self._right_layout.addWidget(self._options_widget)
        self._right_layout.addStretch()
        scroll_widget = QWidget()
        scroll_widget.setLayout(self._right_layout)

        self._right_scroll.setWidget(scroll_widget)
        self._work_area_layout.addWidget(self._right_scroll)

        self.main_layout.addLayout(self._work_area_layout)

        self.main_widget.setLayout(self.main_layout)
        self.setCentralWidget(self.main_widget)

    def closeEvent(self, event: QCloseEvent) -> None:
        """Hide window on close."""
        CONFIG.set_gui_settings("window_geometry", self.saveGeometry().data().hex())
        if CONFIG.get_gui_settings("minimize_to_tray"):
            event.ignore()
            self.hide()
        else:
            super().closeEvent(event)

    def show(self) -> None:
        """Override show."""
        self._load_geometry()
        return super().show()

    def _load_geometry(self) -> None:
        """Load window geometry."""
        geometry = CONFIG.get_gui_settings("window_geometry")
        if geometry:
            self.restoreGeometry(bytes.fromhex(geometry))

    @asyncSlot()
    async def _change_current_pair(self, pair: str) -> None:
        """Change current pair.

        Unsubscribe from current pair and subscribe to new pair. Update chart.

        Args:
            pair (str): Pair name e.g Crypto.BTC/USD.
        """
        # Do nothing if pair is same as current
        if pair == self._current_pair:
            return

        # Disconnect signal to avoid visual glitch
        self._fetcher_message_bus.subscribed_prices_signal.disconnect(
            self._chart.update_chart_tick,
        )
        self._fetcher_message_bus.positions_signal.disconnect(
            self._chart.draw_positions,
        )
        self._fetcher_message_bus.orders_signal.disconnect(self._chart.draw_orders)

        await self._current_exchange.fetcher.unsubscribe_to_price(self._current_pair)
        await self._current_exchange.fetcher.subscribe_to_price(pair)

        self._current_pair = pair
        self._chart.current_pair = pair

        history = await self._current_exchange.fetch_price_history(
            pair,
            self._chart.current_timeframe,
            bars_num=200,
        )
        history_dataframe = pandas.DataFrame(history)
        self._chart.set_start_data(history_dataframe, reset=True)
        minimal_digits = ui_utils.get_minimal_digits(history["low"][0], 4)
        self._chart.main_chart.precision(minimal_digits)

        await self._perps_trade.update_current_pair(pair)

        self._fetcher_message_bus.subscribed_prices_signal.connect(
            self._chart.update_chart_tick,
        )
        self._fetcher_message_bus.positions_signal.connect(
            self._chart.draw_positions,
        )
        self._fetcher_message_bus.orders_signal.connect(self._chart.draw_orders)

    @asyncSlot()
    async def _set_chart_timeframe(self, resolution: str) -> None:
        """Set chart timeframe."""
        history = await self._current_exchange.fetch_price_history(
            self._current_pair,
            resolution,
            bars_num=200,
        )
        history_dataframe = pandas.DataFrame(history)
        self._chart.set_start_data(history_dataframe)

    @asyncSlot()
    async def _fill_news_list(self) -> None:
        """Fill news list with old news."""
        self._news_list.setDisabled(True)
        old_news = await self._news_manager.fetch_old_news(self._news_list.max_news)
        self._news_list.fill_old_news(old_news)
        self._news_list.setDisabled(False)

    @asyncSlot()
    async def change_account(self, account: KeyringAccount) -> None:
        """Chage current account.

        Update exchangeBase on all modules of the list
        _exchange_update_affected and account_update_affected
        with the on_new_exchange and on_new_account.

        Args:
            account (KeyringAccount): Account to switch to.
        """
        toast_id = Toast.show_message(
            "Changing account...",
            type_=ToastType.WARNING,
            timeout=20000,
        )
        self.setEnabled(False)
        new_exchange = await VALID_EXCHANGES[str(account.exchange_name)].create(
            self._fetcher_message_bus,
        )
        await self.on_new_exchange(new_exchange)
        self.setEnabled(True)
        Toast.update_message(
            toast_id,
            "Account changed successfully!",
            ToastType.SUCCESS,
        )

    def _update_quick_trade_values(self) -> None:
        """Update trade values."""
        self._news_list.update_news_trade_buttons()
        self._perps_trade.update_trade_buttons()
        Toast.show_message("Trade values updated!", type_=ToastType.SUCCESS)

    async def on_new_exchange(self, new_exchange: ExchangeBase) -> None:
        """Update exchangeBase on all modules of the list exchange_update_affected."""
        await self._current_exchange.stop_async()

        self._news_message_bus.blockSignals(True)
        self._fetcher_message_bus.blockSignals(True)

        self._current_exchange = await new_exchange.create(self._fetcher_message_bus)

        # Init price fetching loops
        self._async_tasks.append(asyncio.create_task(self._current_exchange.fetch_prices()))
        self._current_pair = self._current_exchange.default_pair

        # Update modules with new exchange
        for module in self._exchange_update_affected:
            LOGGER.debug("Setting up new exchange: %s", module)
            module.blockSignals(True)
            module.on_new_exchange(new_exchange)  # type: ignore
            module.blockSignals(False)

        # Enable options if available
        self._options_widget.setEnabled(self._current_exchange.has_options())
        self._options_widget.setVisible(self._current_exchange.has_options())

        # Rebuild news list
        await self._fill_news_list()

        # Update chart
        await self._set_chart_timeframe("1")

        # reload config from database
        CONFIG.load_config()

        # Update account info
        self._account_info.approve_btn.setVisible(
            not await self._current_exchange.is_ready_to_trade(),
        )

        for module in self._account_update_affected:
            module.blockSignals(True)
            module.on_new_account()  # type: ignore
            module.blockSignals(False)

        self._news_message_bus.blockSignals(False)
        self._fetcher_message_bus.blockSignals(False)

        await self._current_exchange.fetcher.resubscribe_on_going_connections()
        self._update_quick_trade_values()

    async def stop_async(self) -> None:
        """Stop all async tasks and cleanup for deletion."""
        LOGGER.debug("Stopping Plutus Terminal async")
        for task in self._async_tasks:
            task.cancel()
        await asyncio.gather(
            self._news_manager.stop_async(),
            self._current_exchange.stop_async(),
        )



"""Custom types for plutus_terminal."""

from .exchange.types import (
    ExchangeType,
    NewAccountInfo,
    OptionsBuyParams,
    OptionsDirection,
    OptionsDuration,
    OptionsOrdersParams,
    OptionsOrderType,
    OptionsPercent,
    OptionsRisk,
    OptionsSortingBy,
    OptionsSortingDestination,
    OptionsStrategy,
    PerpsPosition,
    PerpsTradeDirection,
    PerpsTradeType,
    PriceData,
    PriceHistory,
)
from .news.filter.types import ActionType, FilterType
from .news.types import NewsData

__all__ = [
    "NewsData",
    "ExchangeType",
    "PerpsTradeType",
    "PerpsTradeDirection",
    "PriceData",
    "PriceHistory",
    "PerpsPosition",
    "OptionsSortingBy",
    "OptionsSortingDestination",
    "OptionsOrderType",
    "OptionsDirection",
    "OptionsDuration",
    "OptionsPercent",
    "OptionsOrdersParams",
    "OptionsRisk",
    "OptionsStrategy",
    "OptionsBuyParams",
    "NewAccountInfo",
    "FilterType",
    "ActionType",
]



"""Utilities for foxify exchange."""

from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, NotRequired, TypedDict

from web3 import AsyncWeb3
from web3.types import Gwei, Nonce, TxParams, Wei

from plutus_terminal.core.exchange.web3 import web3_utils
from plutus_terminal.ui.widgets.toast import Toast, ToastType

if TYPE_CHECKING:
    from decimal import Decimal

    from eth_account.signers.local import LocalAccount
    from eth_typing import ChecksumAddress
    from web3.contract.async_contract import AsyncContract

    from plutus_terminal.core.exchange.types import PerpsTradeType
    from plutus_terminal.core.types_ import PerpsTradeDirection

LOGGER = logging.getLogger(__name__)

REFERRAL_CODE = "0x706c757475735f7465726d696e616c0000000000000000000000000000000000"

STABLE_ADDRESS = AsyncWeb3.to_checksum_address(
    "0xaf88d065e77c8cC2239327C5EDb3A432268e5831",
)

FOXIFY_POSITION_ROUTER = AsyncWeb3.to_checksum_address(
    "0x444A707dF66B124c92E0B9cB86ceb359F2A570FD",
)

FOXIFY_VAULT = AsyncWeb3.to_checksum_address(
    "0xF45CD8739C80cD069D816bf5e60a79FE090CEe30",
)

FOXIFY_ROUTER = AsyncWeb3.to_checksum_address(
    "0xaFc9e58A663Da7eA472ECF31ee5514221B7268b6",
)

FOXIFY_REFERRAL_STORAGE = AsyncWeb3.to_checksum_address(
    "0x7564A44c916485D44822C9728e11d78B2de4C302",
)

FOXIFY_ORDER_BOOK = AsyncWeb3.to_checksum_address(
    "0x72f399b84dBC018AC35af8064501594704a765B9",
)

FOXIFY_ROUTER_CALLBACK = AsyncWeb3.to_checksum_address(
    "0xabDBacB016705D5647C0d9eea6990765DC045120",
)

FOXIFY_FUNDED_FACTORY = AsyncWeb3.to_checksum_address(
    "0x57B0421e3B2A7E18102223F2e1E052b9f8aD1eFB",
)

USDC_DECIMAL_PLACES = 6
MAX_SLIPPAGE = 0.005


class OpenTradingArgs(TypedDict):
    """Args for trading perps."""

    index_token: str
    amount_in: Decimal
    size_delta: Decimal
    trade_direction: PerpsTradeDirection
    acceptable_price: Decimal
    trade_type: PerpsTradeType
    take_profit: Decimal
    stop_loss: Decimal


class CloseTradingArgs(TypedDict):
    """Args for closing trading perps."""

    index_token: str
    size_delta: Decimal
    trade_direction: PerpsTradeDirection
    acceptable_price: Decimal


class ReduceTradingArgs(CloseTradingArgs):
    """Args for reducing trading perps."""

    trade_type: PerpsTradeType


class CancelOrderArgs(TypedDict):
    """Args for canceling order."""

    order_index: int
    reduce_only: bool


class EditOrderArgs(TypedDict):
    """Args for editing order."""

    order_index: int
    size_delta: Decimal
    acceptable_price: Decimal
    trigger_above_threshold: bool
    trade_type: PerpsTradeType


def build_vault_contract(
    web3_provider: AsyncWeb3,
) -> AsyncContract:
    """Build vault contract.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.

    Returns:
        AsyncContract: Contract for vault
    """
    with Path.open(
        Path(__file__).parent.joinpath("abi/foxify_vault.json"),
    ) as f:
        vault_abi = json.load(f)

    return web3_provider.eth.contract(address=FOXIFY_VAULT, abi=vault_abi)


def build_referral_storage_contract(web3_provider: AsyncWeb3) -> AsyncContract:
    """Build referral contract.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.

    Returns:
        AsyncContract: Contract for referral
    """
    with Path.open(
        Path(__file__).parent.joinpath("abi/foxify_referral_storage.json"),
    ) as f:
        referral_abi = json.load(f)

    return web3_provider.eth.contract(address=FOXIFY_REFERRAL_STORAGE, abi=referral_abi)


def build_position_router_contract(
    web3_provider: AsyncWeb3,
) -> AsyncContract:
    """Build position router contract.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.

    Returns:
        AsyncContract: Contract for position router.
    """
    with Path.open(
        Path(__file__).parent.joinpath("abi/foxify_position_router.json"),
    ) as f:
        abi = json.load(f)

    return web3_provider.eth.contract(address=FOXIFY_POSITION_ROUTER, abi=abi)


def build_router_contract(
    web3_provider: AsyncWeb3,
) -> AsyncContract:
    """Build router contract.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.

    Returns:
        AsyncContract: Contract for router.
    """
    with Path.open(
        Path(__file__).parent.joinpath("abi/foxify_router.json"),
    ) as f:
        abi = json.load(f)

    return web3_provider.eth.contract(address=FOXIFY_ROUTER, abi=abi)


def build_order_book_contract(
    web3_provider: AsyncWeb3,
) -> AsyncContract:
    """Build order book contract.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.

    Returns:
        AsyncContract: Contract for order book.
    """
    with Path.open(
        Path(__file__).parent.joinpath("abi/foxify_order_book.json"),
    ) as f:
        abi = json.load(f)

    return web3_provider.eth.contract(address=FOXIFY_ORDER_BOOK, abi=abi)


def build_funded_factory_contract(
    web3_provider: AsyncWeb3,
) -> AsyncContract:
    """Build funded factory contract.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.

    Returns:
        AsyncContract: Contract for order book.
    """
    with Path.open(
        Path(__file__).parent.joinpath("abi/foxify_funded_factory.json"),
    ) as f:
        abi = json.load(f)

    return web3_provider.eth.contract(address=FOXIFY_FUNDED_FACTORY, abi=abi)


def build_funded_trader_contract(
    trader_address: ChecksumAddress,
    web3_provider: AsyncWeb3,
) -> AsyncContract:
    """Build funded trader contract.

    Args:
        trader_address (ChecksumAddress): Trader address.
        web3_provider (AsyncWeb3): Web3 provider.

    Returns:
        AsyncContract: Contract for order book.
    """
    with Path.open(
        Path(__file__).parent.joinpath("abi/foxify_funded_trader.json"),
    ) as f:
        abi = json.load(f)

    return web3_provider.eth.contract(address=trader_address, abi=abi)


async def build_time_lock_contract(
    web3_provider: AsyncWeb3,
) -> AsyncContract:
    """Build time lock contract.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.

    Returns:
        AsyncContract: Contract for time lock.
    """
    with Path.open(
        Path(__file__).parent.joinpath("abi/foxify_time_lock.json"),
    ) as f:
        abi = json.load(f)

    time_lock_address = await build_vault_contract(web3_provider).functions.owner().call()
    return web3_provider.eth.contract(address=time_lock_address, abi=abi)


def build_stable_contract(
    web3_provider: AsyncWeb3,
) -> AsyncContract:
    """Build Options core contract.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.

    Returns:
        AsyncContract: Contract for options core.
    """
    with Path.open(Path(__file__).parent.parent.joinpath("web3/abi/usdc.json")) as f:
        stable_abi = json.load(f)

    return web3_provider.eth.contract(address=STABLE_ADDRESS, abi=stable_abi)


async def is_plugin_approved(
    web3_provider: AsyncWeb3,
    plugin_address: ChecksumAddress,
    wallet_address: ChecksumAddress,
) -> bool:
    """Verify if token plugin is approved on router.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.
        plugin_address (ChecksumAddress): Plugin address.
        wallet_address (ChecksumAddress): Address to check for approval.

    Returns:
        bool: True if the token is already approved. False otherwise.
    """
    router_contract = build_router_contract(web3_provider)

    approved = await router_contract.functions.approvedPlugins(
        wallet_address,
        plugin_address,
    ).call()
    return bool(approved)


async def approve_plugin(
    web3_provider: AsyncWeb3,
    plugin_address: ChecksumAddress,
    web3_account: LocalAccount,
) -> None:
    """Approve plugin.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.
        plugin_address (ChecksumAddress): Plugin address.
        web3_account (LocalAccount): Web3 account.
    """
    if await is_plugin_approved(
        web3_provider,
        plugin_address,
        web3_account.address,
    ):
        return

    router_contract = build_router_contract(web3_provider)
    transaction_count: Nonce = await web3_provider.eth.get_transaction_count(
        web3_account.address,
    )

    transaction_params: TxParams = {
        "gas": Wei(400000),
        "from": web3_account.address,
        "nonce": transaction_count,
        "value": Wei(0),
    }

    approval_transaction = await router_contract.functions.approvePlugin(
        web3_account.address,
        plugin_address,
    ).build_transaction(transaction_params)

    # Estimate gas for transaction with extra gas for speed
    approval_transaction.update(
        {"gas": await web3_utils.estimate_gas(web3_provider, approval_transaction)},
    )

    signed_txn = web3_account.sign_transaction(approval_transaction)
    txn = await web3_provider.eth.send_raw_transaction(signed_txn.rawTransaction)

    await web3_utils.await_receipt_and_report(
        txn,
        web3_provider,
        "Plugin approved.",
        "https://arbiscan.io/tx/",
        LOGGER,
    )


async def is_stable_approved(
    web3_provider: AsyncWeb3,
    spender_address: ChecksumAddress,
    wallet_address: ChecksumAddress,
) -> bool:
    """Verify is token spend is approved on wallet.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.
        spender_address (ChecksumAddress): Spender address.
        wallet_address (ChecksumAddress): Address to check allowance.

    Returns:
        bool: True if the token is already approved. False otherwise.
    """
    stable_contract = build_stable_contract(web3_provider)

    approved = await stable_contract.functions.allowance(
        wallet_address,
        spender_address,
    ).call()
    approved_quantity = await stable_contract.functions.balanceOf(wallet_address).call()
    if int(approved) <= int(approved_quantity):
        return False
    return True


async def approve_stable(
    web3_provider: AsyncWeb3,
    spender_address: ChecksumAddress,
    web3_account: LocalAccount,
) -> None:
    """Approve stable for spender.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.
        spender_address (ChecksumAddress): Spender address.
        web3_account (LocalAccount): Web3 account.
    """
    if await is_stable_approved(web3_provider, spender_address, web3_account.address):
        return
    stable_contract = build_stable_contract(web3_provider)
    transaction_count: Nonce = await web3_provider.eth.get_transaction_count(
        web3_account.address,
    )

    transaction_params: TxParams = {
        "gas": Wei(400000),
        "from": web3_account.address,
        "nonce": transaction_count,
        "value": Wei(0),
        "maxFeePerGas": await web3_utils.estimate_gas_price(web3_provider, Gwei(0)),
        "maxPriorityFeePerGas": web3_provider.to_wei(0, "gwei"),
    }
    max_approval = 115792089237316195423570985008687907853269984665640564039457584007913129639935
    approval_transaction = await stable_contract.functions.approve(
        spender_address,
        max_approval,
    ).build_transaction(transaction_params)

    # Estimate gas for transaction with extra gas for speed
    approval_transaction.update(
        {"gas": await web3_utils.estimate_gas(web3_provider, approval_transaction)},
    )

    signed_txn = web3_account.sign_transaction(approval_transaction)
    txn = await web3_provider.eth.send_raw_transaction(signed_txn.rawTransaction)

    await web3_utils.await_receipt_and_report(
        txn,
        web3_provider,
        "Stable approved.",
        "https://arbiscan.io/tx/",
        LOGGER,
    )


async def ensure_referral(
    web3_provider: AsyncWeb3,
    web3_account: LocalAccount,
) -> None:
    """Ensure referral is set.

    Args:
        web3_provider (AsyncWeb3): Web3 provider.
        web3_account (LocalAccount): Web3 account.
    """
    referral_contract = build_referral_storage_contract(web3_provider)
    current_referral = web3_provider.to_hex(
        await referral_contract.functions.traderReferralCodes(
            web3_account.address,
        ).call(),
    )
    if current_referral == REFERRAL_CODE:
        return

    transaction_count: Nonce = await web3_provider.eth.get_transaction_count(
        web3_account.address,
    )

    transaction_params: TxParams = {
        "gas": Wei(400000),
        "from": web3_account.address,
        "nonce": transaction_count,
        "value": Wei(0),
        "maxFeePerGas": await web3_utils.estimate_gas_price(web3_provider, Gwei(0)),
        "maxPriorityFeePerGas": web3_provider.to_wei(0, "gwei"),
    }
    set_referral_trasaction = await referral_contract.functions.setTraderReferralCodeByUser(
        REFERRAL_CODE,
    ).build_transaction(transaction_params)
    # Estimate gas for transaction with extra gas for speed
    set_referral_trasaction.update(
        {"gas": await web3_utils.estimate_gas(web3_provider, set_referral_trasaction)},
    )
    signed_txn = web3_account.sign_transaction(set_referral_trasaction)
    txn = await web3_provider.eth.send_raw_transaction(signed_txn.rawTransaction)

    await web3_utils.await_receipt_and_report(
        txn,
        web3_provider,
        "Referral set.",
        "https://arbiscan.io/tx/",
        LOGGER,
    )
